/* sane - Scanner Access Now Easy.

   Copyright (C) 2000 Mustek.
   Maintained by Tom Wang <tom.wang@mustek.com.tw>

   Updates (C) 2001 by Henning Meier-Geinitz.

   This file is part of the SANE package.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston,
   MA 02111-1307, USA.

   As a special exception, the authors of SANE give permission for
   additional uses of the libraries contained in this release of SANE.

   The exception is that, if you link a SANE library with other files
   to produce an executable, this does not by itself cause the
   resulting executable to be covered by the GNU General Public
   License.  Your use of that executable is in no way restricted on
   account of linking the SANE library code into it.

   This exception does not, however, invalidate any other reasons why
   the executable file might be covered by the GNU General Public
   License.

   If you submit changes to SANE to the maintainers to be included in
   a subsequent release, you agree by submitting the changes that
   those changes may be distributed with this exception intact.

   If you write modifications of your own for SANE, it is your choice
   whether to permit this exception to apply to your modifications.
   If you do not wish that, delete this exception notice.

   This file implements a SANE backend for Mustek 1200UB flatbed scanners.  */

#include "../include/sane/sane.h"
#include "mustek_usb_low.h"

/* ma1017 ---------------------------------------------------------------*/

#include <fcntl.h>
#include <sys/stat.h>

ma1017 *
usb_low_init (void)
{
  SANE_Int i;
  ma1017 * chip;

  DBG (7, "usb_low_init\n");

  chip = (ma1017 *) malloc (sizeof (ma1017));
  if (!chip)
    {
      DBG (3, "usb_low_init: couldn't malloc %d bytes for chip\n",
	   sizeof (ma1017));
      return NULL;
    }

  /* io */
  chip->isRowing = SANE_FALSE;
  chip->isOpened = SANE_FALSE;
  chip->nUSBPipe = -1;


  /* Construction/Destruction */
  chip->isOpened = SANE_FALSE;
  chip->isRowing = SANE_FALSE;
  
  /* A2 */
  chip->bcmAppend = 0x00;
  chip->bcmTestSRAM = 0x00;
  chip->bcmFixPattern = 0x00;
  /* A4 */
  chip->bcmSelect = 0x00;
  chip->bcmFrontEnd = 0x00;
  /* A6 */
  chip->bcmRGBSelPin = 0x02;
  chip->bcmASICIoPins = 0x9c;
  /* A7 */
  chip->bcmTiming = 0xe8;
  chip->bcmSRAMBank = 0x02;
  /* A8 */
  chip->bcmDummyMSB = 0x00;
  chip->bcmCCDWidthMSB = 0x00;
  chip->bCMTTableLength = 0x00;
  /* A9 */
  chip->bCMTSecondPos = 0x00;
  /* A10 + A8ID5 */
   chip->wCCDWidth = 0x0c80;
  /* A11 + A8ID6 */
   chip->wDummy = 0x0020;
  /* A12 + A13 */
   chip->wByteWidth = 0x09f6;
   /* A14 + A30W */
   chip->wLoopCount = 0x0db5;
   /* A15 */
   chip->bcmMotorEnable = 0x00;
   chip->bcmMotorMovement = 0x60;
   chip->bcmMotorDirection = 0x10;
   chip->bcmMotorSignal = 0x00;
   chip->bcmMotorHome = 0x00;
   /* A16 */
   chip->bcmPixelDeepth = 0x00;
   chip->bcmImageInvert = 0x00;
   chip->bcmOptical600 = 0x00;
   chip->bcmSampleWay = 0x06;
   /* A17 + A18 + A19 */
   chip->bRedRef = 0xff;
   chip->bGreenRef = 0xff;
   chip->bBlueRef = 0xff;
   /* A20 + A21 + A22 */
   chip->bRedPD = 0x00;
   chip->bGreenPD = 0x00;
   chip->bBluePD = 0x00;
   /* A23 */
   chip->bcmA23 = 0x80;
   /* A24 */
   chip->bcmFy1Delay = 0x00;
   chip->bcmSpecialAD = 0x00;
   /* A27 */
   chip->bcmSCLK = 0x00;
   chip->bcmSEN = 0x00;
   chip->bSerialLength = 0x10;
   
   /* Use for Rowing */
   chip->pGetRow = NULL;

   chip->dwCMTTableLength = 0x00000000;
   chip->dwCMTSecondPos = 0x00000000;
   chip->wRowSize = 0x00;
   chip->wSoftResample = 0x01;
   chip->dwTotalLine = 0x00;
   chip->dwLeftLine = 0x00;
   for (i = 0; i < 32; i++)
     chip->isTransferTable[i] = SANE_FALSE;
   chip->bIs300600Switch = SANE_FALSE;

   return chip;
}

void
usb_low_exit (ma1017 * chip)
{
  DBG (7, "usb_low_exit: chip = %p\n", chip);
  if (chip)
    {
      if (chip->nUSBPipe >= 0)
	{
	  close (chip->nUSBPipe);
	  chip->nUSBPipe = -1;
	}
      free (chip);
      chip = 0;
    }
}

/* A0 ~ A1 */
SANE_Bool
usb_low_SetCMTTable (ma1017 * chip, int iIndex, CHANNEL chChannel,
		     SANE_Bool isMoveMotor, SANE_Bool isTransfer)
{
  SANE_Byte bPattern = ((SANE_Byte) iIndex) << 4;
  SANE_Byte bRegNo = 0;
  CONTROLWORD CW_SetCMTTable;

  if (!chip->isOpened)
    {
      DBG (7, "usb_low_SetCMTTable Not Opened yet\n");
      return SANE_FALSE;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_SetCMTTable Stop Rowing first\n");
      return SANE_FALSE;
    }
  if ((unsigned int) iIndex > 31)
    {
      DBG (7, "usb_low_SetCMTTable CMT Index (%d) exceed 31", iIndex);
      return SANE_FALSE;
    }

  switch (chChannel)
    {
    case CH_RED:
      bPattern |= 0x04;
      break;
    case CH_GREEN:
      bPattern |= 0x08;
      break;
    case CH_BLUE:
      bPattern |= 0x0c;
      break;
    default:
      break;
    }
  if (isMoveMotor)
    bPattern |= 0x02;
  if (isTransfer)
    bPattern |= 0x01;
  if (iIndex > 15)
    bRegNo++;
  CW_SetCMTTable.bByte = bPattern;
  CW_SetCMTTable.bRegNo = bRegNo;

  usb_low_io_WriteReg (chip, CW_SetCMTTable);
  chip->isTransferTable[iIndex] = isTransfer;
  return SANE_TRUE;
}

/* A2 */
SANE_Byte usb_low_GetA2 (ma1017 * chip)
{
  SANE_Byte bPattern;

  if (!chip->isOpened)
    {
      DBG (7, "usb_low_GetA2 Not Opened yet\n");
      return 0;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_GetA2 Stop Rowing first\n");
      return 0;
    }

  bPattern = usb_low_io_ReadReg (chip, 2);

  chip->bcmAppend = bPattern & 0x10;
  chip->bcmTestSRAM = bPattern & 0x20;
  chip->bcmFixPattern = bPattern & 0x80;
  return bPattern;
}

SANE_Bool usb_low_StartCMTTable (ma1017 * chip)
{
  SANE_Byte bPattern;
  CONTROLWORD CW_StartCMTTable;

  bPattern = 0x02 | chip->bcmAppend | chip->bcmTestSRAM | chip->bcmFixPattern;
  CW_StartCMTTable.bByte = bPattern;
  CW_StartCMTTable.bRegNo = 2;

  if (!chip->isOpened)
    {
      DBG (7, "usb_low_StartCMTTable Not Opened yet\n");
      return SANE_FALSE;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_StartCMTTable Already Rowing\n");
      return SANE_FALSE;
    }

  usb_low_io_StartRowing (chip, CW_StartCMTTable);
  chip->isRowing = SANE_TRUE;
  return SANE_TRUE;
}

SANE_Byte usb_low_StopCMTTable (ma1017 * chip)
{
  SANE_Byte bPattern;
  CONTROLWORD CW_StopCMTTable;
  SANE_Byte bByte;

  bPattern = 0x01 | chip->bcmAppend | chip->bcmTestSRAM | chip->bcmFixPattern;
  CW_StopCMTTable.bByte = bPattern;
  CW_StopCMTTable.bRegNo = 2;

  if (!chip->isOpened)
    {
      DBG (7, "usb_low_StopCMTTable Not Opened yet\n");
      return 0;
    }
  if (!chip->isRowing)
    {
      DBG (7, "usb_low_StopCMTTable Not Rowing yet\n");
      return 0;
    }

  bByte = usb_low_io_StopRowing (chip, CW_StopCMTTable);
  chip->isRowing = SANE_FALSE;
  return bByte;
}

#if 0
SANE_Bool usb_low_ClearInterrupt (ma1017 * chip)
{
  SANE_Byte bPattern;
  CONTROLWORD CW_ClearInterrupt;

  bPattern = 0x40 | chip->bcmAppend | chip->bcmTestSRAM | chip->bcmFixPattern;
  CW_ClearInterrupt.bByte = bPattern;
  CW_ClearInterrupt.bRegNo = 2;

  if (!chip->isOpened)
    {
      DBG (7, "usb_low_ClearInterrupt Not Opened yet\n");
      return SANE_FALSE;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_ClearInterrupt Stop Rowing first\n");
      return SANE_FALSE;
    }

  usb_low_io_WriteReg (chip, CW_ClearInterrupt);
  return SANE_TRUE;
}
#endif

SANE_Bool usb_low_SetTestSRAMMode (ma1017 * chip, SANE_Bool isTest)
{
  SANE_Byte bPattern;
  CONTROLWORD CW_SetTestSRAMMode;

  bPattern = chip->bcmAppend | chip->bcmTestSRAM | chip->bcmFixPattern;
  CW_SetTestSRAMMode.bByte = bPattern;
  CW_SetTestSRAMMode.bRegNo = 2;

  if (!chip->isOpened)
    {
      DBG (7, "usb_low_SetTestSRAMMode Not Opened yet\n");
      return SANE_FALSE;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_SetTestSRAMMode Stop Rowing first\n");
      return SANE_FALSE;
    }

  if (isTest)
    chip->bcmTestSRAM = 0x20;
  else
    chip->bcmTestSRAM = 0x00;

  usb_low_io_WriteReg (chip, CW_SetTestSRAMMode);
  return SANE_TRUE;
}

#if 0
SANE_Bool usb_low_SoftwareReset (ma1017 * chip)
{
  SANE_Byte bPattern;
  CONTROLWORD CW_SoftwareReset;

 bPattern = 0x08 | chip->bcmAppend | chip->bcmTestSRAM | chip->bcmFixPattern;
 CW_SoftwareReset.bByte = bPattern;
 CW_SoftwareReset.bRegNo = 2;

  if (!chip->isOpened)
    {
      DBG (7, "usb_low_SoftwareReset Not Opened yet\n");
      return SANE_FALSE;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_SoftwareReset Stop Rowing first\n");
      return SANE_FALSE;
    }
  usb_low_io_WriteReg (chip, CW_SoftwareReset);
  return SANE_TRUE;
}
#endif

SANE_Bool usb_low_SetFixPattern (ma1017 * chip, SANE_Bool isFix)
{
  SANE_Byte bPattern;
  CONTROLWORD CW_SetFixPattern;

  bPattern = chip->bcmAppend | chip->bcmTestSRAM | chip->bcmFixPattern;
  CW_SetFixPattern.bByte = bPattern;
  CW_SetFixPattern.bRegNo = 2;

  if (!chip->isOpened)
    {
      DBG (7, "usb_low_SetFixPattern Not Opened yet\n");
      return SANE_FALSE;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_SetFixPattern Stop Rowing first\n");
      return SANE_FALSE;
    }

  if (isFix)
    chip->bcmFixPattern = 0x80;
  else
    chip->bcmFixPattern = 0x00;

  usb_low_io_WriteReg (chip, CW_SetFixPattern);
  return SANE_TRUE;
}

/* A3 */
/*
void usb_low_AdjustTiming(
									TIMING tdFySh1Delay,
									TIMING tdFySh1Width,
									TIMING tdFySh2Delay,
									TIMING tdFyRbDelay,
									TIMING tdFyRbWidth)
{
	if (!chip->isOpened)
		throw XMA1017(EC_ProgrammingError,"usb_low_AdjustTiming Not Opened yet\n");
	if (chip->isRowing)
		throw XMA1017(EC_ProgrammingError,"usb_low_AdjustTiming Stop Rowing first\n");

	SANE_Byte bPattern=0x00; 
	switch (tdFySh1Delay) {
		case 0:
			bPattern|=0x00;
			break;
		case 1:
			bPattern|=0x20;
			break;
		case 2:
			bPattern|=0x10;
			break;
		case 3:
			bPattern|=0x30;
			break;
		default:
			throw XMA1017(EC_InvalidParameter,"usb_low_AdjustTiming tdFySh1Delay Error.\n");
	}
	switch (tdFySh1Width) {
		case 0:
			bPattern|=0x00;
			break;
		case 1:
			bPattern|=0x08;
			break;
		default:
			throw XMA1017(EC_InvalidParameter,"usb_low_AdjustTiming tdFySh1Width Error.\n");
	}
	switch (tdFySh2Delay) {
		case 0:
			bPattern|=0x00;
			break;
		case 1:
			bPattern|=0x80;
			break;
		case 2:
			bPattern|=0x40;
			break;
		case 3:
			bPattern|=0xc0;
			break;
		default:
			throw XMA1017(EC_InvalidParameter,"usb_low_AdjustTiming tdFySh2Delay Error.\n");
	}
	switch (tdFyRbDelay) {
		case 0:
			bPattern|=0x00;
			break;
		case 1:
			bPattern|=0x04;
			break;
		case 2:
			bPattern|=0x02;
			break;
		case 3:
			bPattern|=0x06;
			break;
		default:
			throw XMA1017(EC_InvalidParameter,"usb_low_AdjustTiming tdFyRbDelay Error.\n");
	}
	switch (tdFyRbWidth) {
		case 0:
			bPattern|=0x00;
			break;
		case 1:
			bPattern|=0x01;
			break;
		default:
			throw XMA1017(EC_InvalidParameter,"usb_low_AdjustTiming tdFyRbWidth Error.\n");
	}
	CONTROLWORD CW_AdjustTiming={bPattern,3};

	try {
		usb_low_io_WriteReg(chip,CW_AdjustTiming);
	}
	catch(XMA1017IO x) {
		x.ReportError();
		throw XMA1017(x.ExceptCode(),"usb_low_AdjustTiming Exception Occured.\n");
	}
}
*/

SANE_Bool usb_low_AdjustTiming (ma1017 * chip, SANE_Byte bAdjustPattern)
{
  CONTROLWORD CW_AdjustTiming;

  CW_AdjustTiming.bByte = bAdjustPattern;
  CW_AdjustTiming.bRegNo = 3;

  if (!chip->isOpened)
    {
      DBG (7, "usb_low_AdjustTiming Not Opened yet\n");
      return SANE_FALSE;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_AdjustTiming Stop Rowing first\n");
      return SANE_FALSE;
    }

  usb_low_io_WriteReg (chip, CW_AdjustTiming);
  return SANE_TRUE;
}

/* A4 */
SANE_Byte usb_low_GetA4 (ma1017 * chip)
{
  SANE_Byte bPattern;
  if (!chip->isOpened)
    {
      DBG (7, "usb_low_GetA4 Not Opened yet\n");
      return 0;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_GetA4 Stop Rowing first\n");
      return 0;
    }

  bPattern = usb_low_io_ReadReg (chip, 4);
  chip->bcmSelect = bPattern & 0xfe;
  chip->bcmFrontEnd = bPattern & 0x01;
  return bPattern;
}

/*
void usb_low_SelectTiming(
									SANE_Bool isFyTgHighPulse,
									SANE_Bool isFySh1HighPulse,
									SANE_Bool isFySh2HighPulse,
									SANE_Bool isFyRbHighPulse,
									SANE_Bool isADClockPositive,
									SANE_Bool isOSTNegative,
									SANE_Bool isOSTAndFySh1,
									SANE_Bool isFrontEndEnable)
{
	if (!isOpened)
		throw XMA1017(EC_ProgrammingError,"usb_low_SelectTiming Not Opened yet\n");
	if (isRowing)
		throw XMA1017(EC_ProgrammingError,"usb_low_SelectTiming Stop Rowing first\n");

	SANE_Byte bPattern=0x00;
	if (isFyTgHighPulse)
		bPattern|=0x80;
	if (isFySh1HighPulse)
		bPattern|=0x20;
	if (isFySh2HighPulse)
		bPattern|=0x40;
	if (isFyRbHighPulse)
		bPattern|=0x10;
	if (isADClockPositive)
		bPattern|=0x08;
	if (isOSTNegative)
		bPattern|=0x04;
	if (isOSTAndFySh1)
		bPattern|=0x02;
	if (isFrontEndEnable)
		bPattern|=0x01;
	bcmSelect=bPattern & 0xfe;
	bcmFrontEnd=bPattern & 0x01;
	CONTROLWORD CW_SelectTiming={bPattern,4};

	try {
		usb_low_io_WriteReg(chip,CW_SelectTiming);
	}
	catch(XMA1017IO x) {
		x.ReportError();
		throw XMA1017(x.ExceptCode(),"usb_low_SelectTiming Exception Occured.\n");
	}
}
*/

SANE_Bool usb_low_SelectTiming (ma1017 * chip, SANE_Byte bSelectPattern)
{
  CONTROLWORD CW_SelectTiming;

  CW_SelectTiming.bByte = bSelectPattern;
  CW_SelectTiming.bRegNo = 4;

  if (!chip->isOpened)
    {
      DBG (7, "usb_low_SelectTiming Not Opened yet\n");
      return SANE_FALSE;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_SelectTiming Stop Rowing first\n");
      return SANE_FALSE;
    }

  chip->bcmSelect = bSelectPattern & 0xfe;
  chip->bcmFrontEnd = bSelectPattern & 0x01;

  usb_low_io_WriteReg (chip, CW_SelectTiming);
  return SANE_TRUE;
}

SANE_Bool usb_low_TurnFrontEndMode (ma1017 * chip, SANE_Bool isOn)
{
  CONTROLWORD CW_TurnFrontEndMode;

  if (!chip->isOpened)
    {
      DBG (7, "usb_low_TurnFrontEndMode Not Opened yet\n");
      return SANE_FALSE;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_TurnFrontEndMode Stop Rowing first\n");
      return SANE_FALSE;
    }

  if (isOn)
    chip->bcmFrontEnd = 0x01;
  else
    chip->bcmFrontEnd = 0x00;

  CW_TurnFrontEndMode.bByte = chip->bcmSelect | chip->bcmFrontEnd;
  CW_TurnFrontEndMode.bRegNo = 4;

  usb_low_io_WriteReg (chip, CW_TurnFrontEndMode);
  return SANE_TRUE;
}

/* A6 */
SANE_Byte usb_low_GetA6 (ma1017 * chip)
{
  SANE_Byte bPattern;
  if (!chip->isOpened)
    {
      DBG (7, "usb_low_GetA6 Not Opened yet\n");
      return 0;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_GetA6 Stop Rowing first\n");
      return 0;
    }

  bPattern = usb_low_io_ReadReg (chip, 6);

  chip->bcmASICIoPins = bPattern & 0xdc;
  chip->bcmRGBSelPin = bPattern & 0x03;
  return bPattern;
}

/*
void usb_low_SetASICIoPins(
									 SANE_Bool isIO2Input,
									 SANE_Bool isIO3Input,
									 SANE_Bool isInterruptEnable,
									 SANE_Bool isSTOPInput,
									 SANE_Bool isFy2Input)
{
	if (!isOpened)
		throw XMA1017(EC_ProgrammingError,"usb_low_SetASICIoPins Not Opened yet\n");
	if (isRowing)
		throw XMA1017(EC_ProgrammingError,"usb_low_SetASICIoPins Stop Rowing first\n");

	bcmASICIoPins=0x00;
	if (isIO2Input)
		bcmASICIoPins|=0x04;
	if (isIO3Input)
		bcmASICIoPins|=0x80;
	if (isInterruptEnable)
		bcmASICIoPins|=0x40;
	if (isSTOPInput)
		bcmASICIoPins|=0x10;
	if (isFy2Input)
		bcmASICIoPins|=0x08;
	CONTROLWORD CW_SetASICIoPins={bcmASICIoPins | bcmRGBSelPin, 6};

	try {
		usb_low_io_WriteReg(chip,CW_SetASICIoPins);
	}
	catch(XMA1017IO x) {
		x.ReportError();
		throw XMA1017(x.ExceptCode(),"usb_low_SetASICIoPins Exception Occured.\n");
	}
}
*/

SANE_Bool usb_low_SetASICIoPins (ma1017 * chip, SANE_Byte bIoPinPattern)
{
  CONTROLWORD CW_SetASICIoPins;
  /*={chip->bcmASICIoPins | chip->bcmRGBSelPin, 6}; */

  if (!chip->isOpened)
    {
      DBG (7, "usb_low_SetASICIoPins Not Opened yet\n");
      return SANE_FALSE;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_SetASICIoPins Stop Rowing first\n");
      return SANE_FALSE;
    }

  chip->bcmASICIoPins = bIoPinPattern & 0xdc;

  CW_SetASICIoPins.bByte = chip->bcmASICIoPins | chip->bcmRGBSelPin;
  CW_SetASICIoPins.bRegNo = 6;

  usb_low_io_WriteReg (chip, CW_SetASICIoPins);
  return SANE_TRUE;
}

/*
void usb_low_SetRGBSelPins(
									SANE_Bool isExtendRGBPin,
									SANE_Bool isRGBPinLowActive)
{
	if (!isOpened)
		throw XMA1017(EC_ProgrammingError,"usb_low_SetRGBSelPin Not Opened yet\n");
	if (isRowing)
		throw XMA1017(EC_ProgrammingError,"usb_low_SetRGBSelPin Stop Rowing first\n");

	bcmRGBSelPin=0x00;
	if (isExtendRGBPin)
		bcmRGBSelPin|=0x02;
	if (isRGBPinLowActive)
		bcmRGBSelPin|=0x01;
	CONTROLWORD CW_SetRGBSelPin={bcmASICIoPins | bcmRGBSelPin, 6};

	try {
		usb_low_io_WriteReg(chip,CW_SetRGBSelPin);
	}
	catch(XMA1017IO x) {
		x.ReportError();
		throw XMA1017(x.ExceptCode(),"usb_low_SetRGBSelPin Exception Occured.\n");
	}
}
*/

SANE_Bool usb_low_SetRGBSelPins (ma1017 * chip, SANE_Byte bRGBSelPinPattern)
{
  CONTROLWORD CW_SetRGBSelPin;

  if (!chip->isOpened)
    {
      DBG (7, "usb_low_SetRGBSelPin Not Opened yet\n");
      return SANE_FALSE;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_SetRGBSelPin Stop Rowing first\n");
      return SANE_FALSE;
    }
  chip->bcmRGBSelPin = bRGBSelPinPattern & 0x03;
  CW_SetRGBSelPin.bByte = chip->bcmASICIoPins | chip->bcmRGBSelPin;
  CW_SetRGBSelPin.bRegNo = 6;

  usb_low_io_WriteReg (chip, CW_SetRGBSelPin);
  return SANE_FALSE;
}

/* A7 */
SANE_Byte usb_low_GetA7 (ma1017 * chip)
{
  SANE_Byte bPattern;

  if (!chip->isOpened)
    {
      DBG (7, "usb_low_GetA7 Not Opened yet\n");
      return 0;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_GetA7 Stop Rowing first\n");
      return 0;
    }
  bPattern = usb_low_io_ReadReg (chip, 7);

  chip->bcmTiming = bPattern & 0xfc;
  chip->bcmSRAMBank = bPattern & 0x03;
  return bPattern;
}

/*
void usb_low_SetTiming(
							 SANE_Bool isCISMode,
							 SANE_Bool isCANONMode,
							 SANE_Bool isHighPixelRate,
							 SANE_Bool isShiftFy,
							 SANE_Bool is2PixelPerFy,
							 SANE_Bool isFy1Quarter)
{
	if (!chip->isOpened)
		throw XMA1017(EC_ProgrammingError,"usb_low_SetTiming Not Opened yet\n");
	if (chip->isRowing)
		throw XMA1017(EC_ProgrammingError,"usb_low_SetTiming Stop Rowing first\n");

	chip->bcmTiming=0x00;
	if (isCISMode)
		chip->bcmTiming|=0x80;
	if (isCANONMode)
		chip->bcmTiming|=0x40;
	if (isHighPixelRate)
		chip->bcmTiming|=0x20;
	if (isShiftFy)
		chip->bcmTiming|=0x10;
	if (is2PixelPerFy)
		chip->bcmTiming|=0x08;
	if (isFy1Quarter)
		chip->bcmTiming|=0x04;
	CONTROLWORD CW_SetTiming={bcmTiming | bcmSRAMBank, 7};

	try {
		usb_low_io_WriteReg(chip,CW_SetTiming);
	}
	catch(XMA1017IO x) {
		x.ReportError();
		throw XMA1017(x.ExceptCode(),"usb_low_SetTiming Exception Occured.\n");
	}
}
*/

SANE_Bool usb_low_SetTiming (ma1017 * chip, SANE_Byte bTimingPattern)
{
  CONTROLWORD CW_SetTiming;

  if (!chip->isOpened)
    {
      DBG (7, "usb_low_SetTiming Not Opened yet\n");
      return SANE_FALSE;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_SetTiming Stop Rowing first\n");
      return SANE_FALSE;
    }

  chip->bcmTiming = bTimingPattern & 0xfc;
  CW_SetTiming.bByte = chip->bcmTiming | chip->bcmSRAMBank;
  CW_SetTiming.bRegNo = 7;

  usb_low_io_WriteReg (chip, CW_SetTiming);
  return SANE_TRUE;
}

SANE_Bool usb_low_SetSRAMBank (ma1017 * chip, BANKSIZE bsBankSize)
{
  CONTROLWORD CW_SetSRAMBank;

  if (!chip->isOpened)
    {
      DBG (7, "usb_low_SetSRAMBank Not Opened yet\n");
      return SANE_FALSE;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_SetSRAMBank Stop Rowing first\n");
      return SANE_FALSE;
    }

  switch (bsBankSize)
    {
    case BS_4K:
      chip->bcmSRAMBank = 0x00;
      break;
    case BS_8K:
      chip->bcmSRAMBank = 0x01;
      break;
    case BS_16K:
      chip->bcmSRAMBank = 0x02;
      break;
    default:
      DBG (7, "usb_low_SetSRAMBank bsBankSize Error.\n");
      return SANE_FALSE;
      break;
    }
  CW_SetSRAMBank.bByte = chip->bcmTiming | chip->bcmSRAMBank;
  CW_SetSRAMBank.bRegNo = 7;

  usb_low_io_WriteReg (chip, CW_SetSRAMBank);
  return SANE_TRUE;
}

/* A8 */
SANE_Byte usb_low_GetA8 (ma1017 * chip)
{
  SANE_Byte bPattern;
  if (!chip->isOpened)
    {
      DBG (7, "usb_low_GetA8 Not Opened yet\n");
      return 0;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_GetA8 Stop Rowing first\n");
      return 0;
    }

  bPattern = usb_low_io_ReadReg (chip, 8);

  chip->bcmDummyMSB = bPattern & 0x40;
  chip->bcmCCDWidthMSB = bPattern & 0x20;
  chip->bCMTTableLength = bPattern & 0x1f;
  chip->wCCDWidth =
    ((chip->wCCDWidth / 32) & 0x00ff) * 32 +
    ((chip->bcmCCDWidthMSB == 0) ? 0 : 0x0100 * 32);
  chip->wDummy =
    ((chip->wDummy / 32) & 0x00ff) * 32 +
    ((chip->bcmDummyMSB == 0) ? 0 : 0x0100 * 32);
  return bPattern;
}

SANE_Bool usb_low_SetCMTTableLength (ma1017 * chip, SANE_Byte bLength)
{
  CONTROLWORD CW_SetCMTTableLength;

  if (!chip->isOpened)
    {
      DBG (7, "usb_low_SetCMTTableLength Not Opened yet\n");
      return SANE_FALSE;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_SetCMTTableLength Stop Rowing first\n");
      return SANE_FALSE;
    }
  if (bLength > 32)
    {
      DBG (7, "usb_low_SetCMTTableLength Length: %d exceed 32.",
	      (int) bLength);
      return SANE_FALSE;
    }
  if (bLength == 0)
    {
      DBG (7, "usb_low_SetCMTTableLength Length is equal to 0.\n");
      return SANE_FALSE;
    }

  chip->bCMTTableLength = bLength - 1;
  chip->dwCMTTableLength = (SANE_Word) bLength;
  CW_SetCMTTableLength.bByte =
    chip->bCMTTableLength | chip->bcmCCDWidthMSB |
    chip->bcmDummyMSB;
  CW_SetCMTTableLength.bRegNo = 8;

  usb_low_io_WriteReg (chip, CW_SetCMTTableLength);
  return SANE_TRUE;
}

/* A9 */
SANE_Byte usb_low_GetA9 (ma1017 * chip)
{
  SANE_Byte bPattern;
  if (!chip->isOpened)
    {
      DBG (7, "usb_low_GetA9 Not Opened yet\n");
      return 0;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_GetA9 Stop Rowing first\n");
      return 0;
    }

  bPattern = usb_low_io_ReadReg (chip, 9);
  chip->bCMTSecondPos = bPattern & 0x1f;
  return bPattern;
}

SANE_Bool usb_low_SetCMTSecondPosition (ma1017 * chip, SANE_Byte bPosition)
{
  CONTROLWORD CW_SetCMTSecondPosition;

  if (!chip->isOpened)
    {
      DBG (7, "usb_low_SetCMTSecondPosition Not Opened yet\n");
      return SANE_FALSE;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_SetCMTSecondPosition Stop Rowing first\n");
      return SANE_FALSE;
    }
  if (bPosition > 31)
    {
      DBG (7, "usb_low_SetCMTTableLength Length: %d exceed 31.",
	      (int) bPosition);
      return SANE_FALSE;
    }

  chip->bCMTSecondPos = bPosition;
  chip->dwCMTSecondPos = (SANE_Word) (bPosition);
  CW_SetCMTSecondPosition.bByte = chip->bCMTSecondPos;
  CW_SetCMTSecondPosition.bRegNo = 9;

  usb_low_io_WriteReg (chip, CW_SetCMTSecondPosition);
  return SANE_TRUE;
}

/* A10 + A8ID5 */
SANE_Byte usb_low_GetA10 (ma1017 * chip)
{
  SANE_Byte bA10Pattern;

  if (!chip->isOpened)
    {
      DBG (7, "usb_low_GetA10 Not Opened yet\n");
      return SANE_FALSE;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_GetA10 Stop Rowing first\n");
      return SANE_FALSE;
    }

  bA10Pattern = usb_low_io_ReadReg (chip, 10);
  chip->wCCDWidth =
    ((SANE_Word) (bA10Pattern)) * 32 +
    ((chip->bcmCCDWidthMSB == 0) ? 0 : 0x0100 * 32);
  return bA10Pattern;
}

SANE_Bool usb_low_SetCCDWidth (ma1017 * chip, SANE_Word wCCDWidth)
{
  CONTROLWORD CW_SetCCDWidthMSB;
  CONTROLWORD CW_SetCCDWidth;

  if (!chip->isOpened)
    {
      DBG (7, "usb_low_SetCCDWidth Not Opened yet\n");
      return SANE_FALSE;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_SetCCDWidth Stop Rowing first\n");
      return SANE_FALSE;
    }
  if (wCCDWidth / 32 > 0x01ff)
    {
      DBG (7, "usb_low_SetCCDWidth Width: %d exceed.", (int) wCCDWidth);
      return SANE_FALSE;
    }

  chip->wCCDWidth = wCCDWidth;
  wCCDWidth /= 32;
  if (HIBYTE (wCCDWidth) == 0x01)
    chip->bcmCCDWidthMSB = 0x20;
  else
    chip->bcmCCDWidthMSB = 0x00;
  CW_SetCCDWidthMSB.bByte =
    chip->bCMTTableLength | chip->bcmCCDWidthMSB |
    chip->bcmDummyMSB;
  CW_SetCCDWidthMSB.bRegNo = 8;
  CW_SetCCDWidth.bByte = LOBYTE (wCCDWidth);
  CW_SetCCDWidth.bRegNo = 10;

  usb_low_io_WriteReg (chip, CW_SetCCDWidthMSB);
  usb_low_io_WriteReg (chip, CW_SetCCDWidth);
  return SANE_TRUE;
}

/* A11 + A8ID6 */
SANE_Byte
usb_low_GetA11 (ma1017 * chip)
{
  SANE_Byte bA11Pattern;
  if (!chip->isOpened)
    {
      DBG (7, "usb_low_GetA11 Not Opened yet\n");
      return SANE_FALSE;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_GetA11 Stop Rowing first\n");
      return SANE_FALSE;
    }

  bA11Pattern = usb_low_io_ReadReg (chip, 11);

  chip->wDummy =
    ((SANE_Word) (bA11Pattern)) * 32 +
    ((chip->bcmDummyMSB == 0) ? 0 : 0x0100 * 32);
  return bA11Pattern;
}

SANE_Bool
usb_low_SetDummy (ma1017 * chip, SANE_Word wDummy)
{
  CONTROLWORD CW_SetDummyMSB;
  CONTROLWORD CW_SetDummy;

  if (!chip->isOpened)
    {
      DBG (7, "usb_low_SetDummy Not Opened yet\n");
      return SANE_FALSE;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_SetDummy Stop Rowing first\n");
      return SANE_FALSE;
    }
  if (wDummy / 32 > 0x01ff)
    {
      DBG (7, "usb_low_SetDummy Width: %d exceed.", (int) wDummy);
      return SANE_FALSE;
    }

  chip->wDummy = wDummy;
  wDummy /= 32;
  wDummy++;
  if (HIBYTE (wDummy) == 0x01)
    chip->bcmDummyMSB = 0x40;
  else
    chip->bcmDummyMSB = 0x00;
  CW_SetDummyMSB.bByte =
    chip->bCMTTableLength | chip->bcmCCDWidthMSB |
    chip->bcmDummyMSB;
  CW_SetDummyMSB.bRegNo = 8;
  CW_SetDummy.bByte = LOBYTE (wDummy);
  CW_SetDummy.bRegNo = 11;

  usb_low_io_WriteReg (chip, CW_SetDummyMSB);
  usb_low_io_WriteReg (chip, CW_SetDummy);
  return SANE_TRUE;
}

/* A12 + A13 */
SANE_Byte
usb_low_GetA12 (ma1017 * chip)
{
  SANE_Byte bPattern;
  if (!chip->isOpened)
    {
      DBG (7, "usb_low_GetA12 Not Opened yet\n");
      return 0;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_GetA12 Stop Rowing first\n");
      return 0;
    }

  bPattern = usb_low_io_ReadReg (chip, 12);

  chip->wByteWidth =
    (chip->wByteWidth & 0x3f00) + ((SANE_Word) bPattern);
  chip->wSoftResample =
    (chip->wSoftResample == 0) ? 1 : chip->wSoftResample;
  chip->pGetRow =
    (chip->wSoftResample ==
     1) ? & usb_low_GetRowDirect : & usb_low_GetRowResample;
  chip->wRowSize = chip->wByteWidth / chip->wSoftResample;
  return bPattern;
}

SANE_Byte
usb_low_GetA13 (ma1017 * chip)
{
  SANE_Byte bPattern;

  if (!chip->isOpened)
    {
      DBG (7, "usb_low_GetA13 Not Opened yet\n");
      return 0;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_GetA13 Stop Rowing first\n");
      return 0;
    }

  bPattern = usb_low_io_ReadReg (chip, 13);

  chip->wByteWidth =
    (chip->wByteWidth & 0x00ff) + (((SANE_Word) (bPattern & 0x3f)) << 8);
  chip->wSoftResample =
    (chip->wSoftResample == 0) ? 1 : chip->wSoftResample;
  chip->pGetRow =
    (chip->wSoftResample ==
     1) ? & usb_low_GetRowDirect : & usb_low_GetRowResample;
  chip->wRowSize = chip->wByteWidth / chip->wSoftResample;
  return bPattern;
}

SANE_Bool
usb_low_SetImageByteWidth (ma1017 * chip, SANE_Word wRowSize)
{
  CONTROLWORD CW_SetLOImageByteWidth;
  CONTROLWORD CW_SetHIImageByteWidth;

  if (!chip->isOpened)
    {
      DBG (7, "usb_low_SetImageByteWidth Not Opened yet\n");
      return SANE_FALSE;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_SetImageByteWidth Stop Rowing first\n");
      return SANE_FALSE;
    }

  chip->wRowSize = wRowSize;
  chip->wSoftResample =
    (chip->wSoftResample == 0) ? 1 : chip->wSoftResample;
  chip->pGetRow =
    (chip->wSoftResample ==
     1) ? &usb_low_GetRowDirect : &usb_low_GetRowResample;
  chip->wByteWidth = chip->wRowSize * chip->wSoftResample;
  if (chip->wByteWidth > 0x3fff)
    {
      DBG (7, "usb_low_SetImageByteWidth Width: %d exceed.",
	      (int) chip->wByteWidth);
      return SANE_FALSE;
    }

  CW_SetLOImageByteWidth.bByte = LOBYTE (chip->wByteWidth);
  CW_SetLOImageByteWidth.bRegNo = 12;

  CW_SetHIImageByteWidth.bByte = HIBYTE (chip->wByteWidth);
  CW_SetHIImageByteWidth.bRegNo = 13;

  usb_low_io_WriteReg (chip, CW_SetLOImageByteWidth);
  usb_low_io_WriteReg (chip, CW_SetHIImageByteWidth);
  return SANE_TRUE;
}

SANE_Bool
usb_low_SetSoftResample (ma1017 * chip, SANE_Word wSoftResample)
{
  CONTROLWORD CW_SetLOImageByteWidth;
  CONTROLWORD CW_SetHIImageByteWidth;

  if (!chip->isOpened)
    {
      DBG (7, "usb_low_SetSoftResample Not Opened yet\n");
      return SANE_FALSE;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_SetSoftResample Stop Rowing first\n");
      return SANE_FALSE;
    }
  if (wSoftResample == 0x00)
    {
      DBG (7, "usb_low_SetSoftResample wSoftResample==0 .\n");
      return SANE_FALSE;
    }

  chip->wSoftResample = wSoftResample;
  chip->pGetRow =
    (chip->wSoftResample ==
     1) ? &usb_low_GetRowDirect : &usb_low_GetRowResample;
  chip->wByteWidth = chip->wRowSize * chip->wSoftResample;
  if (chip->wByteWidth > 0x3fff)
    {
      DBG (7, "usb_low_SetSoftResample Width: %d exceed.",
	      (int) chip->wByteWidth);
      return SANE_FALSE;
    }

  CW_SetLOImageByteWidth.bByte = LOBYTE (chip->wByteWidth);
  CW_SetLOImageByteWidth.bRegNo = 12;
  CW_SetHIImageByteWidth.bByte = HIBYTE (chip->wByteWidth);
  CW_SetHIImageByteWidth.bRegNo = 13;

  usb_low_io_WriteReg (chip, CW_SetLOImageByteWidth);
  usb_low_io_WriteReg (chip, CW_SetHIImageByteWidth);
  return SANE_TRUE;
}

/* A14 + A30W */
SANE_Bool
usb_low_SetCMTLoopCount (ma1017 * chip, SANE_Word wLoopCount)
{
  CONTROLWORD CW_SetLOCMTLoopCount;
  CONTROLWORD CW_SetHICMTLoopCount;

  CW_SetLOCMTLoopCount.bByte = LOBYTE (wLoopCount);
  CW_SetLOCMTLoopCount.bRegNo = 14;
  CW_SetHICMTLoopCount.bByte = HIBYTE (wLoopCount);
  CW_SetHICMTLoopCount.bRegNo = 30;

  if (!chip->isOpened)
    {
      DBG (7, "usb_low_SetCMTLoopCount Not Opened yet\n");
      return SANE_FALSE;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_SetCMTLoopCount Stop Rowing first\n");
      return SANE_FALSE;
    }

  chip->wLoopCount = wLoopCount;

  usb_low_io_WriteReg (chip, CW_SetLOCMTLoopCount);
  usb_low_io_WriteReg (chip, CW_SetHICMTLoopCount);
  return SANE_TRUE;
}

/* A15 */
SANE_Byte
usb_low_GetA15 (ma1017 * chip)
{
  SANE_Byte bPattern;
  if (!chip->isOpened)
    {
      DBG (7, "usb_low_GetA15 Not Opened yet\n");
      return 0;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_GetA15 Stop Rowing first\n");
      return 0;
    }

  bPattern = usb_low_io_ReadReg (chip, 15);

  chip->bcmMotorEnable = bPattern & 0x80;
  chip->bcmMotorMovement = bPattern & 0x68;
  chip->bcmMotorDirection = bPattern & 10;
  chip->bcmMotorSignal = bPattern & 0x06;
  chip->bcmMotorHome = bPattern & 0x01;
  return bPattern;
}

#if 0
void 
usb_low_SetMotorMode (SANE_Bool isEnable, SANE_Bool isFullStep,
		      SANE_Bool isDoublePhase, SANE_Bool isBackward,
		      SANE_Bool isTwoStep, SANE_Bool isPulseSignal,
		      SANE_Bool isSyncWithImageWrite, SANE_Bool isHome)
{
  if (!isOpened)
    throw XMA1017(EC_ProgrammingError,"usb_low_SetMotorMode Not Opened yet\n");
  if (isRowing)
    throw XMA1017(EC_ProgrammingError,"usb_low_SetMotorMode Stop Rowing first\n");
  
  bcmMotorEnable=0x00;
  bcmMotorMovement=0x00;
  bcmMotorDirection=0x00;
  bcmMotorSignal=0x00;
  bcmMotorHome=0x00;
  if (isEnable)
    bcmMotorEnable|=0x80;
  if (isFullStep)
    bcmMotorMovement|=0x40;
  if (isDoublePhase)
    bcmMotorMovement|=0x20;
  if (isBackward)
    bcmMotorDirection|=0x10;
  if (isTwoStep)
    bcmMotorMovement|=0x08;
  if (isPulseSignal)
    bcmMotorSignal|=0x04;
  if (isSyncWithImageWrite)
    bcmMotorSignal|=0x02;
  if (isHome)
    bcmMotorHome|=0x01;
  CONTROLWORD CW_SetMotorMode={bcmMotorEnable|
			       bcmMotorMovement|
			       bcmMotorDirection|
			       bcmMotorSignal|
			       bcmMotorHome, 15};
  
  try {
    usb_low_io_WriteReg(chip,CW_SetMotorMode);
  }
  catch(XMA1017IO x) {
    x.ReportError();
    throw XMA1017(x.ExceptCode(),"usb_low_SetMotorMode Exception Occured.\n");
  }
}
#endif

#if 0
SANE_Bool
usb_low_SetMotorMode (ma1017 * chip, SANE_Byte bMotorPattern)
{
  CONTROLWORD CW_SetMotorMode;

  CW_SetMotorMode.bByte = bMotorPattern;
  CW_SetMotorMode.bRegNo = 15;

  if (!chip->isOpened)
    {
      DBG (7, "usb_low_SetMotorMode Not Opened yet\n");
      return SANE_FALSE;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_SetMotorMode Stop Rowing first\n");
      return SANE_FALSE;
    }

  chip->bcmMotorEnable = bMotorPattern & 0x80;
  chip->bcmMotorMovement = bMotorPattern & 0x68;
  chip->bcmMotorDirection = bMotorPattern & 10;
  chip->bcmMotorSignal = bMotorPattern & 0x06;
  chip->bcmMotorHome = bMotorPattern & 0x01;

  usb_low_io_WriteReg (chip, CW_SetMotorMode);
  return SANE_TRUE;
}
#endif

SANE_Bool
usb_low_EnableMotor (ma1017 * chip, SANE_Bool isEnable)
{
  CONTROLWORD CW_SetMotorMode;

  if (!chip->isOpened)
    {
      DBG (7, "usb_low_EnableMotor Not Opened yet\n");
      return SANE_FALSE;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_EnableMotor Stop Rowing first\n");
      return SANE_FALSE;
    }

  chip->bcmMotorEnable = 0x00;
  if (isEnable)
    chip->bcmMotorEnable |= 0x80;
  CW_SetMotorMode.bByte =
    chip->bcmMotorEnable | chip->bcmMotorMovement |
    chip->bcmMotorDirection | chip->bcmMotorSignal |
    chip->bcmMotorHome;
  CW_SetMotorMode.bRegNo = 15;

  usb_low_io_WriteReg (chip, CW_SetMotorMode);
  return SANE_TRUE;
}

SANE_Bool
usb_low_SetMotorMovement (ma1017 * chip, SANE_Bool isFullStep,
			   SANE_Bool isDoublePhase, SANE_Bool isTwoStep)
{
  CONTROLWORD CW_SetMotorMode;

  if (!chip->isOpened)
    {
      DBG (7, "usb_low_SetMotorMovement Not Opened yet\n");
      return SANE_FALSE;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_SetMotorMovement Stop Rowing first\n");
      return SANE_FALSE;
    }

  chip->bcmMotorMovement = 0x00;
  if (isFullStep)
    chip->bcmMotorMovement |= 0x40;
  if (isDoublePhase)
    chip->bcmMotorMovement |= 0x20;
  if (isTwoStep)
    chip->bcmMotorMovement |= 0x08;
  CW_SetMotorMode.bByte =
    chip->bcmMotorEnable | chip->bcmMotorMovement |
    chip->bcmMotorDirection | chip->bcmMotorSignal |
    chip->bcmMotorHome;
  CW_SetMotorMode.bRegNo = 15;

  usb_low_io_WriteReg (chip, CW_SetMotorMode);
  return SANE_TRUE;
}

SANE_Bool
usb_low_SetMotorDirection (ma1017 * chip, SANE_Bool isBackward)
{
  CONTROLWORD CW_SetMotorMode;

  if (!chip->isOpened)
    {
      DBG (7, "usb_low_SetMotorDirection Not Opened yet\n");
      return SANE_FALSE;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_SetMotorDirection Stop Rowing first\n");
      return SANE_FALSE;
    }

  chip->bcmMotorDirection = 0x00;
  if (isBackward)
    chip->bcmMotorDirection |= 0x10;
  CW_SetMotorMode.bByte =
    chip->bcmMotorEnable | chip->bcmMotorMovement |
    chip->bcmMotorDirection | chip->bcmMotorSignal |
    chip->bcmMotorHome;
  CW_SetMotorMode.bRegNo = 15;

  usb_low_io_WriteReg (chip, CW_SetMotorMode);
  return SANE_TRUE;
}

SANE_Bool
usb_low_SetMotorSignal (ma1017 * chip, SANE_Byte bSignal)
{
  CONTROLWORD CW_SetMotorSignal;

  if (!chip->isOpened)
    {
      DBG (7, "usb_low_SetMotorSignal Not Opened yet\n");
      return SANE_FALSE;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_SetMotorSignal Stop Rowing first\n");
      return SANE_FALSE;
    }

  chip->bcmMotorSignal = bSignal & 0x06;
  CW_SetMotorSignal.bByte =
    chip->bcmMotorEnable | chip->bcmMotorMovement |
    chip->bcmMotorDirection | chip->bcmMotorSignal |
    chip->bcmMotorHome;
  CW_SetMotorSignal.bRegNo = 15;

  usb_low_io_WriteReg (chip, CW_SetMotorSignal);
  return SANE_TRUE;
}

SANE_Bool
usb_low_MoveMotorHome (ma1017 * chip, SANE_Bool isHome, SANE_Bool isBackward)
{
  CONTROLWORD CW_SetMotorMode;

  if (!chip->isOpened)
    {
      DBG (7, "usb_low_MoveMotorHome: Not Opened yet\n");
      return SANE_FALSE;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_MoveMotorHome: Stop Rowing first\n");
      return SANE_FALSE;
    }

  chip->bcmMotorEnable = 0x00;
  chip->bcmMotorDirection = 0x00;
  chip->bcmMotorHome = 0x00;
  if (isBackward)
    chip->bcmMotorDirection |= 0x10;
  if (isHome)
    {
      chip->bcmMotorEnable |= 0x80;
      chip->bcmMotorHome |= 0x01;
    }
  CW_SetMotorMode.bByte =
    chip->bcmMotorEnable | chip->bcmMotorMovement |
    chip->bcmMotorDirection | chip->bcmMotorSignal |
    chip->bcmMotorHome;
  CW_SetMotorMode.bRegNo = 15;

  usb_low_io_WriteReg (chip, CW_SetMotorMode);
  return SANE_TRUE;
}

/* A16 */
SANE_Byte
usb_low_GetA16 (ma1017 * chip)
{
  SANE_Byte bPattern;
  if (!chip->isOpened)
    {
      DBG (7, "usb_low_GetA16 Not Opened yet\n");
      return 0;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_GetA16 Stop Rowing first\n");
      return 0;
    }

  bPattern = usb_low_io_ReadReg (chip, 16);

  chip->bcmPixelDeepth = bPattern & 0xe0;
  chip->bcmImageInvert = bPattern & 0x10;
  chip->bcmOptical600 = bPattern & 0x08;
  chip->bcmSampleWay = bPattern & 0x07;
  return bPattern;
}

/*
void usb_low_SetImageMode(
									PIXELDEEPTH pdPixelDeepth,
									SANE_Bool isInver,
									SANE_Bool isOptical600,
									SAMPLEWAY swSampleWay)
{
	if (!isOpened)
		throw XMA1017(EC_ProgrammingError,"usb_low_SetImageMode Not Opened yet\n");
	if (isRowing)
		throw XMA1017(EC_ProgrammingError,"usb_low_SetImageMode Stop Rowing first\n");

	bcmPixelDeepth=0x00;
	bcmImageInvert=0x00;
	bcmOptical600=0x00;
	bcmSampleWay=0x00;
	switch (pdPixelDeepth) {
		case PD_1BIT:
			bcmPixelDeepth=0x80;
			break;
		case PD_4BIT:
			bcmPixelDeepth=0xc0;
			break;
		case PD_8BIT:
			bcmPixelDeepth=0x00;
			break;
		case PD_12BIT:
			bcmPixelDeepth=0x20;
			break;
		default:
			throw XMA1017(EC_InvalidParameter,"usb_low_SetImageMode pdPixelDeepth Error.\n");
	}
	if (isInver)
		bcmImageInvert|=0x10;
	if (isOptical600)
		bcmOptical600|=0x08;
	switch (swSampleWay) {
		case SW_P1P6:
			bcmSampleWay=0x01;
			break;
		case SW_P2P6:
			bcmSampleWay=0x02;
			break;
		case SW_P3P6:
			bcmSampleWay=0x03;
			break;
		case SW_P4P6:
			bcmSampleWay=0x04;
			break;
		case SW_P5P6:
			bcmSampleWay=0x05;
			break;
		case SW_P6P6:
			bcmSampleWay=0x06;
			break;
		default:
			throw XMA1017(EC_InvalidParameter,"usb_low_SetImageMode swSampleWay Error.\n");
	}
	CONTROLWORD CW_SetImageMode={bcmPixelDeepth|
		bcmImageInvert|
		bcmOptical600|
		bcmSampleWay, 16};

	try {
		usb_low_io_WriteReg(chip,CW_SetImageMode);
	}
	catch(XMA1017IO x) {
		x.ReportError();
		throw XMA1017(x.ExceptCode(),"usb_low_SetImageMode Exception Occured.\n");
	}
}
*/

#if 0
SANE_Bool
usb_low_SetImageMode (ma1017 * chip, SANE_Byte bImageModePattern)
{
  CONTROLWORD CW_SetImageMode;

  CW_SetImageMode.bByte = bImageModePattern;
  CW_SetImageMode.bRegNo = 16;

  if (!chip->isOpened)
    {
      DBG (7, "usb_low_SetImageMode Not Opened yet\n");
      return SANE_FALSE;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_SetImageMode Stop Rowing first\n");
      return SANE_FALSE;
    }

  chip->bcmPixelDeepth = bImageModePattern & 0xe0;
  chip->bcmImageInvert = bImageModePattern & 0x10;
  chip->bcmOptical600 = bImageModePattern & 0x08;
  chip->bcmSampleWay = bImageModePattern & 0x07;

  usb_low_io_WriteReg (chip, CW_SetImageMode);
  return SANE_TRUE;
}
#endif

SANE_Bool
usb_low_SetImageDpi (ma1017 * chip, SANE_Bool isOptical600,
		      SAMPLEWAY swSampleWay)
{
  CONTROLWORD CW_SetImageMode;

  if (!chip->isOpened)
    {
      DBG (7, "usb_low_SetImageMode Not Opened yet\n");
      return SANE_FALSE;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_SetImageMode Stop Rowing first\n");
      return SANE_FALSE;
    }

  chip->bcmOptical600 = 0x00;
  chip->bcmSampleWay = 0x00;
  if (isOptical600)
    chip->bcmOptical600 |= 0x08;
  switch (swSampleWay)
    {
    case SW_P1P6:
      chip->bcmSampleWay = 0x01;
      break;
    case SW_P2P6:
      chip->bcmSampleWay = 0x02;
      break;
    case SW_P3P6:
      chip->bcmSampleWay = 0x03;
      break;
    case SW_P4P6:
      chip->bcmSampleWay = 0x04;
      break;
    case SW_P5P6:
      chip->bcmSampleWay = 0x05;
      break;
    case SW_P6P6:
      chip->bcmSampleWay = 0x06;
      break;
    default:
      DBG (7, "usb_low_SetImageMode swSampleWay Error.\n");
      return SANE_FALSE;
      break;
    }
  CW_SetImageMode.bByte =
    chip->bcmPixelDeepth | chip->bcmImageInvert |
    chip->bcmOptical600 | chip->bcmSampleWay;
  CW_SetImageMode.bRegNo = 16;

  usb_low_io_WriteReg (chip, CW_SetImageMode);
  return SANE_TRUE;
}

SANE_Bool
usb_low_SetPixelDeepth (ma1017 * chip, PIXELDEEPTH pdPixelDeepth)
{
  CONTROLWORD CW_SetImageMode;

  if (!chip->isOpened)
    {
      DBG (7, "usb_low_SetPixelDeepth Not Opened yet\n");
      return SANE_FALSE;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_SetPixelDeepth Stop Rowing first\n");
      return SANE_FALSE;
    }

  chip->bcmPixelDeepth = 0x00;
  switch (pdPixelDeepth)
    {
    case PD_1BIT:
      chip->bcmPixelDeepth = 0x80;
      break;
    case PD_4BIT:
      chip->bcmPixelDeepth = 0xc0;
      break;
    case PD_8BIT:
      chip->bcmPixelDeepth = 0x00;
      break;
    case PD_12BIT:
      chip->bcmPixelDeepth = 0x20;
      break;
    default:
      DBG (7, "usb_low_SetPixelDeepth pdPixelDeepth Error.\n");
      return SANE_FALSE;
    }
  CW_SetImageMode.bByte =
    chip->bcmPixelDeepth | chip->bcmImageInvert |
    chip->bcmOptical600 | chip->bcmSampleWay;
  CW_SetImageMode.bRegNo = 16;

  usb_low_io_WriteReg (chip, CW_SetImageMode);
  return SANE_TRUE;
}

SANE_Bool
usb_low_InvertImage (ma1017 * chip, SANE_Bool isInvert)
{
  CONTROLWORD CW_SetImageMode;

  if (!chip->isOpened)
    {
      DBG (7, "usb_low_InvertImage Not Opened yet\n");
      return SANE_FALSE;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_InvertImage Stop Rowing first\n");
      return SANE_FALSE;
    }

  chip->bcmImageInvert = 0x00;
  if (isInvert)
    chip->bcmImageInvert |= 0x10;
  CW_SetImageMode.bByte =
    chip->bcmPixelDeepth | chip->bcmImageInvert |
    chip->bcmOptical600 | chip->bcmSampleWay;
  CW_SetImageMode.bRegNo = 16;

  usb_low_io_WriteReg (chip, CW_SetImageMode);
  return SANE_TRUE;
}

#if 0
SANE_Bool
usb_low_SetOptical600 (ma1017 * chip, SANE_Bool isOptical600)
{
  CONTROLWORD CW_SetImageMode;

  if (!chip->isOpened)
    {
      DBG (7, "usb_low_SetOptical600 Not Opened yet\n");
      return SANE_FALSE;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_SetOptical600 Stop Rowing first\n");
      return SANE_FALSE;
    }

  chip->bcmOptical600 = 0x00;
  if (isOptical600)
    chip->bcmOptical600 |= 0x08;
  CW_SetImageMode.bByte =
    chip->bcmPixelDeepth | chip->bcmImageInvert |
    chip->bcmOptical600 | chip->bcmSampleWay;
  CW_SetImageMode.bRegNo = 16;

  usb_low_io_WriteReg (chip, CW_SetImageMode);
  return SANE_TRUE;
}
#endif

#if 0
SANE_Bool
usb_low_SetSampleWay (ma1017 * chip, SAMPLEWAY swSampleWay)
{
  CONTROLWORD CW_SetImageMode;

  if (!chip->isOpened)
    {
      DBG (7, "usb_low_SetSampleWay Not Opened yet\n");
      return SANE_FALSE;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_SetSampleWay Stop Rowing first\n");
      return SANE_FALSE;
    }

  chip->bcmSampleWay = 0x00;
  switch (swSampleWay)
    {
    case SW_P1P6:
      chip->bcmSampleWay = 0x01;
      break;
    case SW_P2P6:
      chip->bcmSampleWay = 0x02;
      break;
    case SW_P3P6:
      chip->bcmSampleWay = 0x03;
      break;
    case SW_P4P6:
      chip->bcmSampleWay = 0x04;
      break;
    case SW_P5P6:
      chip->bcmSampleWay = 0x05;
      break;
    case SW_P6P6:
      chip->bcmSampleWay = 0x06;
      break;
    default:
      DBG (7, "usb_low_SetSampleWay swSampleWay Error.\n");
      return SANE_FALSE;
      break;
    }
  CW_SetImageMode.bByte =
    chip->bcmPixelDeepth | chip->bcmImageInvert |
    chip->bcmOptical600 | chip->bcmSampleWay;
  CW_SetImageMode.bRegNo = 16;

  usb_low_io_WriteReg (chip, CW_SetImageMode);
  return SANE_TRUE;
}
#endif

/* A17 + A18 + A19 */
SANE_Byte
usb_low_GetA17 (ma1017 * chip)
{
  SANE_Byte bPattern;
  if (!chip->isOpened)
    {
      DBG (7, "usb_low_GetA17 Not Opened yet\n");
      return 0;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_GetA17 Stop Rowing first\n");
      return 0;
    }

  bPattern = usb_low_io_ReadReg (chip, 17);

  chip->bRedRef = bPattern;
  return bPattern;
}

SANE_Byte
usb_low_GetA18 (ma1017 * chip)
{
  SANE_Byte bPattern;
  if (!chip->isOpened)
    {
      DBG (7, "usb_low_GetA18 Not Opened yet\n");
      return 0;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_GetA18 Stop Rowing first\n");
      return 0;
    }

  bPattern = usb_low_io_ReadReg (chip, 18);

  chip->bGreenRef = bPattern;
  return bPattern;
}

SANE_Byte
usb_low_GetA19 (ma1017 * chip)
{
  SANE_Byte bPattern;
  if (!chip->isOpened)
    {
      DBG (7, "usb_low_GetA19 Not Opened yet\n");
      return 0;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_GetA19 Stop Rowing first\n");
      return 0;
    }

  bPattern = usb_low_io_ReadReg (chip, 19);

  chip->bBlueRef = bPattern;
  return bPattern;
}

#if 0
SANE_Bool
usb_low_SetReference (ma1017 * chip, SANE_Byte RedRef, SANE_Byte GreenRef,
		       SANE_Byte BlueRef)
{
  CONTROLWORD CS_SetReference[3];
  
  CS_SetReference[0].bByte = RedRef;
  CS_SetReference[0].bRegNo = 17;
  CS_SetReference[1].bByte = GreenRef;
  CS_SetReference[1].bRegNo = 18;
  CS_SetReference[2].bByte = BlueRef;
  CS_SetReference[2].bRegNo = 19;

  if (!chip->isOpened)
    {
      DBG (7, "usb_low_SetReference Not Opened yet\n");
      return SANE_FALSE;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_SetReference Stop Rowing first\n");
      return SANE_FALSE;
    }

  chip->bRedRef = RedRef;
  chip->bGreenRef = GreenRef;
  chip->bBlueRef = BlueRef;

  usb_low_io_WriteReg (chip, CS_SetReference[0]);
  usb_low_io_WriteReg (chip, CS_SetReference[1]);
  usb_low_io_WriteReg (chip, CS_SetReference[2]);
  return SANE_TRUE;
}
#endif

SANE_Bool
usb_low_SetRedRef (ma1017 * chip, SANE_Byte RedRef)
{
  CONTROLWORD CW_SetRedReference;

  CW_SetRedReference.bByte = RedRef;
  CW_SetRedReference.bRegNo = 17;

  if (!chip->isOpened)
    {
      DBG (7, "usb_low_SetRedRef Not Opened yet\n");
      return SANE_FALSE;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_SetRedRef Stop Rowing first\n");
      return SANE_FALSE;
    }

  chip->bRedRef = RedRef;

  usb_low_io_WriteReg (chip, CW_SetRedReference);
  return SANE_TRUE;
}

SANE_Bool
usb_low_SetGreenRef (ma1017 * chip, SANE_Byte GreenRef)
{
  CONTROLWORD CW_SetGreenReference;

  CW_SetGreenReference.bByte = GreenRef;
  CW_SetGreenReference.bRegNo = 18;

  if (!chip->isOpened)
    {
      DBG (7, "usb_low_GreenRef Not Opened yet\n");
      return SANE_FALSE;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_GreenRef Stop Rowing first\n");
      return SANE_FALSE;
    }

  chip->bGreenRef = GreenRef;

  usb_low_io_WriteReg (chip, CW_SetGreenReference);
  return SANE_TRUE;
}

SANE_Bool
usb_low_SetBlueRef (ma1017 * chip, SANE_Byte BlueRef)
{
  CONTROLWORD CW_SetBlueReference;

  CW_SetBlueReference.bByte = BlueRef;
  CW_SetBlueReference.bRegNo = 19;

  if (!chip->isOpened)
    {
      DBG (7, "usb_low_BlueRef Not Opened yet\n");
      return SANE_FALSE;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_BlueRef Stop Rowing first\n");
      return SANE_FALSE;
    }

  chip->bBlueRef = BlueRef;

  usb_low_io_WriteReg (chip, CW_SetBlueReference);
  return SANE_TRUE;
}

/* A20 + A21 + A22 */
SANE_Byte
usb_low_GetA20 (ma1017 * chip)
{
  SANE_Byte bPattern;

  if (!chip->isOpened)
    {
      DBG (7, "usb_low_GetA20 Not Opened yet\n");
      return 0;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_GetA20 Stop Rowing first\n");
      return 0;
    }
  bPattern = usb_low_io_ReadReg (chip, 20);

  chip->bRedPD = bPattern;
  return bPattern;
}

SANE_Byte
usb_low_GetA21 (ma1017 * chip)
{
  SANE_Byte bPattern;

  if (!chip->isOpened)
    {
      DBG (7, "usb_low_GetA21 Not Opened yet\n");
      return 0;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_GetA21 Stop Rowing first\n");
      return 0;
    }

  bPattern = usb_low_io_ReadReg (chip, 21);
  chip->bGreenPD = bPattern;
  return bPattern;
}

SANE_Byte
usb_low_GetA22 (ma1017 * chip)
{
  SANE_Byte bPattern;

  if (!chip->isOpened)
    {
      DBG (7, "usb_low_GetA22 Not Opened yet\n");
      return 0;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_GetA22 Stop Rowing first\n");
      return 0;
    }

  bPattern = usb_low_io_ReadReg (chip, 22);

  chip->bBluePD = bPattern;
  return bPattern;
}

#if 0
SANE_Bool
usb_low_SetPowerDelay (ma1017 * chip, SANE_Byte RedPD, SANE_Byte GreenPD,
			SANE_Byte BluePD)
{
  CONTROLWORD CS_SetPowerDelay[3];

  CS_SetPowerDelay[0].bByte = RedPD;
  CS_SetPowerDelay[0].bRegNo = 20;
  CS_SetPowerDelay[1].bByte = GreenPD;
  CS_SetPowerDelay[1].bRegNo = 21;
  CS_SetPowerDelay[2].bByte = BluePD;
  CS_SetPowerDelay[2].bRegNo = 22;

  if (!chip->isOpened)
    {
      DBG (7, "usb_low_SetPowerDelay Not Opened yet\n");
      return SANE_FALSE;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_SetPowerDelay Stop Rowing first\n");
      return SANE_FALSE;
    }

  chip->bRedPD = RedPD;
  chip->bGreenPD = GreenPD;
  chip->bBluePD = BluePD;

  usb_low_io_WriteReg (chip, CS_SetPowerDelay[0]);
  usb_low_io_WriteReg (chip, CS_SetPowerDelay[1]);
  usb_low_io_WriteReg (chip, CS_SetPowerDelay[2]);

  return SANE_TRUE;
}
#endif

SANE_Bool
usb_low_SetRedPD (ma1017 * chip, SANE_Byte RedPD)
{
  CONTROLWORD CW_SetRedPowerDelay;

  if (!chip->isOpened)
    {
      DBG (7, "usb_low_SetRedPD Not Opened yet\n");
      return SANE_FALSE;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_SetRedPD Stop Rowing first\n");
      return SANE_FALSE;
    }

  chip->bRedPD = RedPD;
  CW_SetRedPowerDelay.bByte = chip->bRedPD;
  CW_SetRedPowerDelay.bRegNo = 20;

  usb_low_io_WriteReg (chip, CW_SetRedPowerDelay);
  return SANE_TRUE;
}

SANE_Bool
usb_low_SetGreenPD (ma1017 * chip, SANE_Byte GreenPD)
{
  CONTROLWORD CW_SetGreenPowerDelay;

  if (!chip->isOpened)
    {
      DBG (7, "usb_low_SetGreenPD Not Opened yet\n");
      return SANE_FALSE;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_SetGreenPD Stop Rowing first\n");
      return SANE_FALSE;
    }

  chip->bGreenPD = GreenPD;
  CW_SetGreenPowerDelay.bByte = chip->bGreenPD;
  CW_SetGreenPowerDelay.bRegNo = 21;

  usb_low_io_WriteReg (chip, CW_SetGreenPowerDelay);
  return SANE_TRUE;
}

SANE_Bool
usb_low_SetBluePD (ma1017 * chip, SANE_Byte BluePD)
{
  CONTROLWORD CW_SetBluePowerDelay;

  if (!chip->isOpened)
    {
      DBG (7, "usb_low_SetBluePD Not Opened yet\n");
      return SANE_FALSE;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_SetBluePD Stop Rowing first\n");
      return SANE_FALSE;
    }

  chip->bBluePD = BluePD;
  CW_SetBluePowerDelay.bByte = chip->bBluePD;
  CW_SetBluePowerDelay.bRegNo = 22;

  usb_low_io_WriteReg (chip, CW_SetBluePowerDelay);
  return SANE_TRUE;
}

/* A23 */
SANE_Byte
usb_low_GetA23 (ma1017 * chip)
{
  SANE_Byte bPattern;

  if (!chip->isOpened)
    {
      DBG (7, "usb_low_GetA23 Not Opened yet\n");
      return 0;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_GetA23 Stop Rowing first\n");
      return 0;
    }

  bPattern = usb_low_io_ReadReg (chip, 23);
  chip->bcmA23 = bPattern;
  return bPattern;
}

SANE_Bool
usb_low_TurnPeripheralPower (ma1017 * chip, SANE_Bool isOn)
{
  CONTROLWORD CW_TurnPeripheralPower;

  if (!chip->isOpened)
    {
      DBG (7, "usb_low_TurnPeripheralPower Not Opened yet\n");
      return SANE_FALSE;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_TurnPeripheralPower Stop Rowing first\n");
      return SANE_FALSE;
    }

  chip->bcmA23 &= 0x7f;
  if (isOn)
    chip->bcmA23 |= 0x80;
  CW_TurnPeripheralPower.bByte = chip->bcmA23;
  CW_TurnPeripheralPower.bRegNo = 23;

  usb_low_io_WriteReg (chip, CW_TurnPeripheralPower);
  return SANE_TRUE;
}

SANE_Bool
usb_low_TurnLampPower (ma1017 * chip, SANE_Bool isOn)
{
  CONTROLWORD CW_TurnLampPower;
  if (!chip->isOpened)
    {
      DBG (7, "usb_low_TurnLampPower Not Opened yet\n");
      return SANE_FALSE;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_TurnLampPower Stop Rowing first\n");
      return SANE_FALSE;
    }

  chip->bcmA23 &= 0xbf;
  if (isOn)
    chip->bcmA23 |= 0x40;
  CW_TurnLampPower.bByte = chip->bcmA23;
  CW_TurnLampPower.bRegNo = 23;

  usb_low_io_WriteReg (chip, CW_TurnLampPower);
  return SANE_TRUE;
}

#if 0
SANE_Bool
usb_low_SetSTOP (ma1017 * chip, SANE_Bool isHigh)
{
  CONTROLWORD CW_SetSTOP;
  if (!chip->isOpened)
    {
      DBG (7, "usb_low_SetSTOP Not Opened yet\n\n");
      return SANE_FALSE;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_SetSTOP Stop Rowing first\n");
      return SANE_FALSE;
    }

  chip->bcmA23 &= 0xdf;
  if (isHigh)
    chip->bcmA23 |= 0x20;
  CW_SetSTOP.bByte = chip->bcmA23;
  CW_SetSTOP.bRegNo = 23;

  usb_low_io_WriteReg (chip, CW_SetSTOP);
  return SANE_TRUE;
}
#endif

#if 0
SANE_Bool
usb_low_SetIO2 (ma1017 * chip, SANE_Bool isHigh)
{
  CONTROLWORD CW_SetIO2;
  if (!chip->isOpened)
    {
      DBG (7, "usb_low_SetIO2 Not Opened yet\n");
      return SANE_FALSE;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_SetIO2 Stop Rowing first\n");
      return SANE_FALSE;
    }

  chip->bcmA23 &= 0xef;
  if (isHigh)
    chip->bcmA23 |= 0x10;
  CW_SetIO2.bByte = chip->bcmA23;
  CW_SetIO2.bRegNo = 23;

  usb_low_io_WriteReg (chip, CW_SetIO2);
  return SANE_TRUE;
}
#endif

SANE_Bool
usb_low_SetIO3 (ma1017 * chip, SANE_Bool isHigh)
{
  CONTROLWORD CW_SetIO3;

  if (!chip->isOpened)
    {
      DBG (7, "usb_low_SetIO3 Not Opened yet\n");
      return SANE_FALSE;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_SetIO3 Stop Rowing first\n");
      return SANE_FALSE;
    }

  chip->bcmA23 &= 0xf7;
  if (isHigh)
    chip->bcmA23 |= 0x08;
  CW_SetIO3.bByte = chip->bcmA23;
  CW_SetIO3.bRegNo = 23;
  usb_low_io_WriteReg (chip, CW_SetIO3);
  return SANE_TRUE;
}

#if 0
SANE_Bool
usb_low_GetPeripheralPower (ma1017 * chip)
{
  if (!chip->isOpened)
    {
      DBG (7, "usb_low_GetPeripheralPower Not Opened yet\n");
      return SANE_FALSE;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_GetPeripheralPower Stop Rowing first\n");
      return SANE_FALSE;
    }

  chip->bcmA23 = usb_low_io_ReadReg (chip, 23);
  return SANE_TRUE;
}
#endif

#if 0
SANE_Bool
usb_low_GetSTOP (ma1017 * chip)
{
  if (!chip->isOpened)
    {
      DBG (7, "usb_low_SetCMTTable Not Opened yet\n");
      return SANE_FALSE;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_SetCMTTable Stop Rowing first\n");
      return SANE_FALSE;
    }

  chip->bcmA23 = usb_low_io_ReadReg (chip, 23);
  return ((chip->bcmA23 & 0x20) != 0x00);
}
#endif
#if 0
SANE_Bool
usb_low_GetIO2 (ma1017 * chip)
{
  if (!chip->isOpened)
    {
      DBG (7, "usb_low_GetIO2 not open\n");
      return SANE_FALSE;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_SetCMTTable Stop Rowing first\n");
      return SANE_FALSE;
    }

  chip->bcmA23 = usb_low_io_ReadReg (chip, 23);
  return ((chip->bcmA23 & 0x10) != 0x00);
}

SANE_Bool
usb_low_GetIO3 (ma1017 * chip)
{
  if (!chip->isOpened)
    {
      DBG (7, "usb_low_SetCMTTable Not Opened yet\n");
      return SANE_FALSE;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_SetCMTTable Stop Rowing first\n");
      return SANE_FALSE;
    }

  chip->bcmA23 = usb_low_io_ReadReg (chip, 23);
  return ((chip->bcmA23 & 0x08) != 0x00);
}

SANE_Bool
usb_low_SetLEDFlashSpeed (ma1017 * chip, FLASHSPEED fsFlashSpeed)
{
  CONTROLWORD CW_SetLEDFlashSpeed;

  if (!chip->isOpened)
    {
      DBG (7, "usb_low_SetLEDFlashSpeed Not Opened yet\n\n");
      return SANE_FALSE;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_SetLEDFlashSpeed Stop Rowing first\n\n");
      return SANE_FALSE;
    }

  chip->bcmA23 &= 0xf9;
  switch (fsFlashSpeed)
    {
    case FS_LOWEST:
      chip->bcmA23 |= 0x00;
      break;
    case FS_LOW:
      chip->bcmA23 |= 0x02;
      break;
    case FS_HIGH:
      chip->bcmA23 |= 0x04;
      break;
    case FS_HIGHEST:
      chip->bcmA23 |= 0x06;
      break;
    default:
      DBG (7, "usb_low_SetLEDFlashSpeed fsFlashSpeed Error.\n");
      return SANE_FALSE;
    }
  CW_SetLEDFlashSpeed.bByte = chip->bcmA23;
  CW_SetLEDFlashSpeed.bRegNo = 23;
  usb_low_io_WriteReg (chip, CW_SetLEDFlashSpeed);
  return SANE_TRUE;
}
#endif

SANE_Bool
usb_low_SetLEDLightAll (ma1017 * chip, SANE_Bool isLightAll)
{
  CONTROLWORD CW_SetLEDLightAll;
  if (!chip->isOpened)
    {
      DBG (7, "usb_low_SetLEDLightAll Not Opened yet\n");
      return SANE_FALSE;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_SetLEDLightAll Stop Rowing first\n");
      return SANE_FALSE;
    }

  chip->bcmA23 &= 0xfe;
  if (isLightAll)
    chip->bcmA23 |= 0x01;
  CW_SetLEDLightAll.bByte = chip->bcmA23;
  CW_SetLEDLightAll.bRegNo = 23;
  usb_low_io_WriteReg (chip, CW_SetLEDLightAll);
  return SANE_TRUE;
}

/* A24 */
SANE_Byte
usb_low_GetA24 (ma1017 * chip)
{
  SANE_Byte bPattern;

  if (!chip->isOpened)
    {
      DBG (7, "usb_low_GetA24 Not Opened yet\n");
      return 0;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_GetA24 Stop Rowing first\n");
      return 0;
    }

  bPattern = usb_low_io_ReadReg (chip, 24);

  chip->bcmFy1Delay = bPattern & 0x01;
  chip->bcmSpecialAD = bPattern & 0x02;
  return bPattern;
}

SANE_Bool
usb_low_SetADTiming (ma1017 * chip, SANE_Byte bPattern)
{
  CONTROLWORD CW_SetADTiming;

  if (!chip->isOpened)
    {
      DBG (7, "usb_low_SetADTiming Not Opened yet\n");
      return SANE_FALSE;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_SetADTiming Stop Rowing first\n");
      return SANE_FALSE;
    }

  chip->bcmFy1Delay = bPattern & 0x01;
  chip->bcmSpecialAD = bPattern & 0x02;

  CW_SetADTiming.bByte = chip->bcmSpecialAD | chip->bcmFy1Delay;
  CW_SetADTiming.bRegNo = 24;

  usb_low_io_WriteReg (chip, CW_SetADTiming);
  return SANE_TRUE;
}

#if 0
SANE_Bool
usb_low_SetFy1Delay (ma1017 * chip, SANE_Bool isFy1Delay)
{
  CONTROLWORD CW_SetFy1Delay;

  if (!chip->isOpened)
    {
      DBG (7, "usb_low_SetFy1Delay Not Opened yet\n");
      return SANE_FALSE;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_SetFy1Delay Stop Rowing first\n");
      return SANE_FALSE;
    }

  if (isFy1Delay)
    chip->bcmFy1Delay = 0x01;
  else
    chip->bcmFy1Delay = 0x00;

  CW_SetFy1Delay.bByte = chip->bcmSpecialAD | chip->bcmFy1Delay;
  CW_SetFy1Delay.bRegNo = 24;

  usb_low_io_WriteReg (chip, CW_SetFy1Delay);
  return SANE_TRUE;
}

SANE_Bool
usb_low_SetSpecialAD (ma1017 * chip, SANE_Bool isSpecialAD)
{
  CONTROLWORD CW_SetSpecialAD;
  /*={chip->bcmSpecialAD | chip->bcmFy1Delay, 24}; */

  if (!chip->isOpened)
    {
      DBG (7, "usb_low_SetSpecialAD Not Opened yet\n");
      return SANE_FALSE;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_SetSpecialAD Stop Rowing first\n");
      return SANE_FALSE;
    }

  if (isSpecialAD)
    chip->bcmSpecialAD = 0x02;
  else
    chip->bcmSpecialAD = 0x00;
  CW_SetSpecialAD.bByte = chip->bcmSpecialAD | chip->bcmFy1Delay;
  CW_SetSpecialAD.bRegNo = 24;


  usb_low_io_WriteReg (chip, CW_SetSpecialAD);
  return SANE_FALSE;
}
#endif

/* A25 + A26 */
SANE_Bool
usb_low_SetSerialByte1 (ma1017 * chip, SANE_Byte Byte1)
{
  CONTROLWORD CW_SetSerialByte1;

  CW_SetSerialByte1.bByte = Byte1;
  CW_SetSerialByte1.bRegNo = 25;

  if (!chip->isOpened)
    {
      DBG (7, "usb_low_SetSerialByte1 Not Opened\n");
      return SANE_FALSE;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_SetSerialByte1 Stop Rowing first\n");
      return SANE_FALSE;
    }

  usb_low_io_WriteReg (chip, CW_SetSerialByte1);
  return SANE_FALSE;
}

SANE_Bool
usb_low_SetSerialByte2 (ma1017 * chip, SANE_Byte Byte2)
{
  CONTROLWORD CW_SetSerialByte2;

  CW_SetSerialByte2.bByte = Byte2;
  CW_SetSerialByte2.bRegNo = 26;

  if (!chip->isOpened)
    {
      DBG (7, "usb_low_SetSerialByte2 Not Opened yet\n");
      return SANE_FALSE;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_SetSerialByte2 Stop Rowing first\n");
      return SANE_FALSE;
    }

  usb_low_io_WriteReg (chip, CW_SetSerialByte2);

  return SANE_TRUE;
}

#if 0
SANE_Bool
usb_low_SendSerial (ma1017 * chip, SANE_Word Word)
{
  CONTROLWORD CS_SendSerial1;
  CONTROLWORD CS_SendSerial2;

  CS_SendSerial1.bByte = HIBYTE (Word);
  CS_SendSerial1.bRegNo = 25;
  CS_SendSerial2.bByte = LOBYTE (Word);
  CS_SendSerial2.bRegNo = 26;

  if (!chip->isOpened)
    {
      DBG (7, "usb_low_SendSerial Not Opened yet\n");
      return SANE_FALSE;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_SendSerial Stop Rowing first\n");
      return SANE_FALSE;
    }

  usb_low_io_WriteReg (chip, CS_SendSerial1);
  usb_low_io_WriteReg (chip, CS_SendSerial2);

  return SANE_TRUE;
}
#endif

/* A27 */
SANE_Byte
usb_low_GetA27 (ma1017 * chip)
{
  SANE_Byte bPattern;

  if (!chip->isOpened)
    {
      DBG (7, "usb_low_GetA27 Not Opened yet\n");
      return 0x0;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_GetA27 Stop Rowing first\n");
      return 0x0;
    }

  bPattern = usb_low_io_ReadReg (chip, 27);

  chip->bcmSCLK = bPattern & 0x80;
  chip->bcmSEN = bPattern & 0x40;
  chip->bSerialLength = bPattern & 0x1f;
  return bPattern;
}

SANE_Bool
usb_low_SetSerialFormat (ma1017 * chip, SANE_Byte Format)
{
  CONTROLWORD CW_SetSerialFormat;

  CW_SetSerialFormat.bByte = Format;
  CW_SetSerialFormat.bRegNo = 27;

  if (!chip->isOpened)
    {
      DBG (7, "usb_low_SetSerialFormat Not Opened yet\n");
      return SANE_FALSE;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_SetSerialFormat Stop Rowing first\n");
      return SANE_FALSE;
    }

  chip->bcmSCLK = Format & 0x80;
  chip->bcmSEN = Format & 0x40;
  chip->bSerialLength = Format & 0x1f;

  return usb_low_io_WriteReg (chip, CW_SetSerialFormat);
}

#if 0
SANE_Bool
usb_low_SetSerialClock (ma1017 * chip, SANE_Bool isRising)
{
  CONTROLWORD CW_SetSerialClock;	/*= */

  if (!chip->isOpened)
    {
      DBG (7, "usb_low_SetSerialClock Not Opened yet\n");
      return SANE_FALSE;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_SetSerialClock Stop Rowing first\n");
      return SANE_FALSE;
    }

  if (isRising)
    chip->bcmSCLK = 0x80;
  else
    chip->bcmSCLK = 0x00;

  CW_SetSerialClock.bByte =
    chip->bcmSCLK | chip->bcmSEN | chip->bSerialLength;
  CW_SetSerialClock.bRegNo = 27;

  return usb_low_io_WriteReg (chip, CW_SetSerialClock);
}

SANE_Bool
usb_low_SetSerialActive (ma1017 * chip, SANE_Bool isHighActive)
{
  CONTROLWORD CW_SetSerialActive;
  /*={chip->bcmSCLK|chip->bcmSEN|chip->bSerialLength,27}; */

  if (!chip->isOpened)
    {
      DBG (7, "usb_low_SetSerialActive Not Opened yet\n");
      return SANE_FALSE;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_SetSerialActive Stop Rowing first\n");
      return SANE_FALSE;
    }

  if (isHighActive)
    chip->bcmSEN = 0x40;
  else
    chip->bcmSEN = 0x00;

  CW_SetSerialActive.bByte =
    chip->bcmSCLK | chip->bcmSEN | chip->bSerialLength;
  CW_SetSerialActive.bRegNo = 27;

  return usb_low_io_WriteReg (chip, CW_SetSerialActive);
}

SANE_Bool
usb_low_SetSerialLength (ma1017 * chip, SANE_Byte bLength)
{
  CONTROLWORD CW_SetSerialLength;

  if (!chip->isOpened)
    {
      DBG (7, "usb_low_SetSerialLength Not Opened yet\n");
      return SANE_FALSE;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_SetSerialLength Stop Rowing first\n");
      return SANE_FALSE;
    }
  if (bLength > 16)
    {
      DBG (7, "usb_low_SetSerialLength bLength>16\n");
      return SANE_FALSE;
    }

  chip->bSerialLength = bLength;
  CW_SetSerialLength.bByte =
    chip->bcmSCLK | chip->bcmSEN | chip->bSerialLength;
  CW_SetSerialLength.bRegNo = 27;


  return usb_low_io_WriteReg (chip, CW_SetSerialLength);
}

/* A29 */
SANE_Bool
usb_low_WriteSRAM (ma1017 * chip, SANE_Byte bPattern)
{
  CONTROLWORD CW_WriteSRAM;

  CW_WriteSRAM.bByte = bPattern;
  CW_WriteSRAM.bRegNo = 29;

  if (!chip->isOpened)
    {
      DBG (7, "usb_low_WriteSRAM Not Opened yet\n");
      return SANE_FALSE;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_WriteSRAM Stop Rowing first\n");
      return SANE_FALSE;
    }

  usb_low_io_WriteReg (chip, CW_WriteSRAM);
  return SANE_TRUE;
}

SANE_Byte
usb_low_GetSRAM (ma1017 * chip)
{
  if (!chip->isOpened)
    {
      DBG (7, "usb_low_GetSRAM Not Opened yet\n");
      return 0;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_GetSRAM Stop Rowing first\n");
      return 0;
    }

  return usb_low_io_ReadReg (chip, 29);
}

/* A30 */
SANE_Byte
usb_low_GetCMTStatus (ma1017 * chip)
{
  if (!chip->isOpened)
    {
      DBG (7, "usb_low_GetCMTStatus Not Opened yet\n");
      return 0;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_GetCMTStatus Stop Rowing first\n");
      return 0;
    }

  return usb_low_io_ReadReg (chip, 30);
}

/* A31 */
SANE_Byte
usb_low_GetInterruptStatus (ma1017 * chip)
{
  if (!chip->isOpened)
    {
      DBG (7, "usb_low_GetInterruptStatus Not Opened yet\n");
      return 0;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_GetInterruptStatus Stop Rowing first\n");
      return 0;
    }

  return usb_low_io_ReadReg (chip, 31);
}
#endif

SANE_Bool
usb_low_GetHomeSensor (ma1017 * chip)
{
  SANE_Byte Byte;

  if (!chip->isOpened)
    {
      DBG (7, "usb_low_GetInterruptStatus Not Opened yet\n");
      return SANE_TRUE;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_GetInterruptStatus Stop Rowing first\n");
      return SANE_TRUE;
    }

  Byte = usb_low_io_ReadReg (chip, 31);
  return (Byte & 0x80) != 0;
}

/* Special Mode */
SANE_Bool
usb_low_StartRowing (ma1017 * chip)
{
  SANE_Word dwLineOfFirst = 0;
  SANE_Word dwLineOfSecond = 0;
  int i;

  if (chip->wLoopCount == 0)
    {
      DBG (7, "usb_low_StartRowing wLoopCount hasn't set yet.\n");
      return SANE_FALSE;
    }
  if (chip->dwCMTTableLength == 0)
    {
      DBG (7, "usb_low_StartRowing dwCMTTableLength hasn't set yet.\n");
      return SANE_FALSE;
    }
  if (chip->dwCMTTableLength <= chip->dwCMTSecondPos)
    {
      DBG (7, "usb_low_StartRowing dwCMTSecondPos cannot larger than "
	   "dwCMTTableLength.\n");
      return SANE_FALSE;
    }

  for (i = 0; i < (int) chip->dwCMTSecondPos; i++)
    {
      if (chip->isTransferTable[i])
	dwLineOfFirst++;
    }
  for (; i < (int) chip->dwCMTTableLength; i++)
    {
      if (chip->isTransferTable[i])
	{
	  dwLineOfFirst++;
	  dwLineOfSecond++;
	}
    }

  chip->dwTotalLine =
    ((SANE_Word) (chip->wLoopCount - 1)) * dwLineOfSecond + dwLineOfFirst;
  chip->dwLeftLine = chip->dwTotalLine;

  usb_low_StartCMTTable (chip);
  return SANE_TRUE;
}

SANE_Byte
usb_low_StopRowing (ma1017 * chip)
{
  return usb_low_StopCMTTable (chip);
}

SANE_Byte
usb_low_WaitRowingStop (ma1017 * chip)
{
  SANE_Byte bByte;
  if (chip->dwTotalLine != 0)
    {
      DBG (7, "usb_low_WaitRowingStop dwTotalLine must be 0.\n");
      return 0;
    }

  bByte = usb_low_io_WaitRowingStop (chip);
  chip->isRowing = SANE_FALSE;
  return bByte;
}

void
usb_low_MatchVariable (ma1017 * chip)
{
  usb_low_GetA2 (chip);
  usb_low_GetA4 (chip);
  usb_low_GetA6 (chip);
  usb_low_GetA7 (chip);
  usb_low_GetA8 (chip);
  usb_low_GetA9 (chip);
  usb_low_GetA10 (chip);
  usb_low_GetA11 (chip);
  usb_low_GetA12 (chip);
  usb_low_GetA13 (chip);
  usb_low_GetA15 (chip);
  usb_low_GetA16 (chip);
  usb_low_GetA17 (chip);
  usb_low_GetA18 (chip);
  usb_low_GetA19 (chip);
  usb_low_GetA20 (chip);
  usb_low_GetA21 (chip);
  usb_low_GetA22 (chip);
  usb_low_GetA23 (chip);
  usb_low_GetA24 (chip);
  usb_low_GetA27 (chip);
}

SANE_Bool
usb_low_Open (ma1017 * chip, SANE_String_Const pDeviceName,
	      SANE_Byte ScannerID)
{
  DBG (7, "usb_low_Open:  ScannerID=%d\n", ScannerID);

  if (chip->isRowing)
    {
      DBG (7, "usb_low_Open Already Rowing\n");
      return SANE_FALSE;
    }
  if (chip->isOpened)
    {
      DBG (7, "usb_low_Open Already Opened.");
      return SANE_FALSE;
    }

  /* usb_low_io_Open("1200 UB","",""); */
  if (!usb_low_io_Open (chip, pDeviceName, "", ""))
    {
      DBG (7, "usb_low_Open Exception Occured.\n");
      return SANE_FALSE;
    }
  chip->isOpened = SANE_TRUE;
  usb_low_MatchVariable (chip);
  return SANE_TRUE;
}

SANE_Bool
usb_low_Close (ma1017 * chip)
{
  if (!chip->isOpened)
    {
      DBG (7, "usb_low_Close Not Opened yet\n");
      return SANE_FALSE;
    }

  if (chip->isRowing)
    usb_low_StopRowing (chip);
  usb_low_io_Close (chip);
  chip->isOpened = SANE_FALSE;
  return SANE_TRUE;
}

/* Modified by Eden */
SANE_Bool
usb_low_SetCIS (ma1017 * chip, SANE_Bool bIs300600Switch)
{
  chip->bIs300600Switch = bIs300600Switch;
  return SANE_TRUE;
}

/* */


SANE_Word
usb_low_GetRow (ma1017 * chip, SANE_Byte * pData)
{
  return (*chip->pGetRow) (chip, pData);
}

SANE_Word
usb_low_GetRowDirect (ma1017 * chip, SANE_Byte * pData)
{
  if (chip->dwLeftLine == 0)
    {
      DBG (7, "usb_low_GetRow dwLeftLine == 0.\n");
      return 0;
    }

  if (chip->dwLeftLine <= 1)
    {
      usb_low_io_ReadRows (chip, pData,
			   chip->wByteWidth);
      usb_low_io_WaitRowingStop (chip);
      chip->dwLeftLine = 0x00;
      chip->isRowing = SANE_FALSE;
      return 0x00;
    }
  else
    {
      if (usb_low_io_ReadRows
	  (chip, pData, chip->wByteWidth))
	{
	  chip->dwLeftLine--;
	}
      return chip->dwLeftLine;
    }
}

SANE_Word
usb_low_GetRowResample (ma1017 * chip, SANE_Byte * pData)
{
  static SANE_Byte s_ResampleBuffer[8 * 1024];
  SANE_Word *pPixelTemp;
  SANE_Word i;
  SANE_Word j;
  SANE_Word k;

  if (chip->dwLeftLine == 0)
    {
      DBG (7, "usb_low_GetRow dwLeftLine == 0.\n");
      return 0;
    }

  if (chip->dwLeftLine <= 1)
    {
      usb_low_io_ReadRows (chip, s_ResampleBuffer,
			   chip->wByteWidth);

      if ((chip->bIs300600Switch == SANE_FALSE)
	  && (chip->bcmPixelDeepth == 0x20))
	{

	  pPixelTemp = (SANE_Word *) malloc (6 * 1024 * sizeof (SANE_Word));

	  j = 0;
	  for (i = 0; i < chip->wByteWidth; i += 3)
	    {
	      pPixelTemp[j] = (SANE_Word) s_ResampleBuffer[i];
	      pPixelTemp[j] |= ((SANE_Word) 
				(s_ResampleBuffer[i + 1] & 0xf0)) << 4;
	      j++;
	      pPixelTemp[j] = ((SANE_Word)
			       (s_ResampleBuffer[i + 1] & 0x0f)) << 8;
	      pPixelTemp[j] |= (SANE_Word) s_ResampleBuffer[i + 2];
	      j++;
	    }

	  k = 0;
	  for (i = 0; i < j; i += chip->wSoftResample * 2)
	    {
	      pData[k] = (SANE_Byte) (pPixelTemp[i] & 0x00ff);
	      k++;
	      pData[k] = (SANE_Byte) ((pPixelTemp[i] & 0x0f00) >> 4);
	      pData[k] |= (SANE_Byte) ((pPixelTemp[i + 2] & 0x0f00) >> 8);
	      k++;
	      pData[k] = (SANE_Byte) (pPixelTemp[i + 2] & 0x00ff);
	      k++;
	    }
	  free (pPixelTemp);
	}
      else
	{
	  for (i = 0; i < chip->wByteWidth;
	       i += chip->wSoftResample)
	    *(pData++) = s_ResampleBuffer[i];
	}
      usb_low_io_WaitRowingStop (chip);
      chip->dwLeftLine = 0x00;
      chip->isRowing = SANE_FALSE;
      return 0x00;
    }
  else
    {
      usb_low_io_ReadRows (chip, s_ResampleBuffer,
			   chip->wByteWidth);

      if ((chip->bIs300600Switch == SANE_FALSE)
	  && (chip->bcmPixelDeepth == 0x20))
	{

	  pPixelTemp = (SANE_Word *) malloc (6 * 1024 * sizeof (SANE_Word));
	  j = 0;
	  for (i = 0; i < chip->wByteWidth; i += 3)
	    {
	      pPixelTemp[j] = (SANE_Word) s_ResampleBuffer[i];
	      pPixelTemp[j] |= ((SANE_Word) (s_ResampleBuffer[i + 1] & 0xf0)) << 4;
	      j++;
	      pPixelTemp[j] = ((SANE_Word) (s_ResampleBuffer[i + 1] & 0x0f)) << 8;
	      pPixelTemp[j] |= (SANE_Word) s_ResampleBuffer[i + 2];
	      j++;
	    }

	  k = 0;
	  for (i = 0; i < j; i += chip->wSoftResample * 2)
	    {
	      pData[k] = (SANE_Byte) (pPixelTemp[i] & 0x00ff);
	      k++;
	      pData[k] = (SANE_Byte) ((pPixelTemp[i] & 0x0f00) >> 4);
	      pData[k] |= (SANE_Byte) ((pPixelTemp[i + 2] & 0x0f00) >> 8);
	      k++;
	      pData[k] = (SANE_Byte) (pPixelTemp[i + 2] & 0x00ff);
	      k++;
	    }
	  free (pPixelTemp);
	}
      else
	{
	  for (i = 0; i < chip->wByteWidth;
	       i += chip->wSoftResample)
	    *(pData++) = s_ResampleBuffer[i];
	}
      chip->dwLeftLine--;
      return chip->dwLeftLine;
    }
}


/* ma1017io -------------------------------------------------------------*/


SANE_Bool
usb_low_io_Open (ma1017 * chip, SANE_String_Const pDeviceName,
		 SANE_String_Const pControlPipe, SANE_String_Const pDataPipe)
{
  DBG (7, "usb_low_io_Open: pControlPipe=%p, pDataPipe=%p\n",
       pControlPipe, pDataPipe);

  if (chip->isOpened)
    {
      DBG (7, "usb_low_io_Open Already Opened\n");
      return SANE_FALSE;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_io_WriteReg Rowing, Stop first.\n");
      return SANE_FALSE;
    }
  if (chip->nUSBPipe)
    close (chip->nUSBPipe);

  if ((chip->nUSBPipe = open ((SANE_String_Const) pDeviceName, O_RDWR)) >= 0)
    {
      chip->isOpened = SANE_TRUE;
      return SANE_TRUE;
    }
  return SANE_FALSE;
}

void
usb_low_io_Close (ma1017 * chip)
{
  if (chip->isOpened)
    {
      if (chip->nUSBPipe)
	{
	  close (chip->nUSBPipe);
	  chip->nUSBPipe = -1;
	}
    }
  chip->isOpened = SANE_FALSE;
  chip->isRowing = SANE_FALSE;
}

SANE_Bool
usb_low_io_StartRowing (ma1017 * chip, CONTROLWORD Control)
{
  ssize_t n;
  if (!chip->isOpened)
    {
      DBG (7, "usb_low_io_StartRowing Open First.\n");
      return SANE_FALSE;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_io_StartRowing Already Rowing.\n");
      return SANE_FALSE;
    }
  Control.bRegNo |= 0x60;
  n = write (chip->nUSBPipe, (void *) &Control, sizeof (CONTROLWORD));
  if (n != sizeof (CONTROLWORD))
    {
      DBG (7, "usb_low_io_StartRowing Failed\n");
      return SANE_FALSE;
    }
  chip->isRowing = SANE_TRUE;
  return SANE_TRUE;
}

SANE_Byte
usb_low_io_StopRowing (ma1017 * chip, CONTROLWORD Control)
{
  SANE_Byte ReadByte;
  ssize_t n;

  if (!chip->isOpened)
    {
      DBG (7, "usb_low_io_StopRowing Open First.\n");
      return 0xff;
    }
  Control.bRegNo |= 0x80;
  n = write (chip->nUSBPipe, (void *) &Control, sizeof (CONTROLWORD));
  if (n != sizeof (CONTROLWORD))
    {
      DBG (7, "usb_low_io_StopRowing Failed\n");
      return 0xff;
    }
  n = read (chip->nUSBPipe, (void *) &ReadByte, 1);
  if (n < 0)
    {
      DBG (7, "usb_low_io_StopRowing Failed\n");
      return 0xff;
    }
  chip->isRowing = SANE_FALSE;
  return ReadByte;
}

SANE_Byte
usb_low_io_WaitRowingStop (ma1017 * chip)
{
  SANE_Byte ReadByte;
  ssize_t n;
  if (!chip->isOpened)
    {
      DBG (7, "usb_low_io_WaitRowingStop Open First.\n");
      return 0;
    }
  if (!chip->isRowing)
    {
      DBG (7, "usb_low_io_WaintRowingStop not Rowing\n");
      return 0;
    }
  n = read (chip->nUSBPipe, (void *) &ReadByte, 1);
  if (n != 1)
    {
      DBG (7, "usb_low_io_WaitRowingStop Failed.\n");
      return 0;
    }
  chip->isRowing = SANE_FALSE;
  return ReadByte;
}

SANE_Bool
usb_low_io_WriteReg (ma1017 * chip, CONTROLWORD Control)
{
  ssize_t n;
  if (!chip->isOpened)
    {
      DBG (7, "usb_low_io_WriteReg Open First.\n");
      return SANE_FALSE;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_io_WriteReg Rowing, Stop first.");
      return SANE_FALSE;
    }
  if (Control.bRegNo > 0x20)
    {
      DBG (7, "usb_low_io_WriteReg RegNo out of range.\n");
      return SANE_FALSE;
    }

  n = write (chip->nUSBPipe, (void *) &Control, sizeof (CONTROLWORD));
  if (n != sizeof (CONTROLWORD))
    {
      DBG (7, "usb_low_io_WriteReg WriteCommand Error\n");
      return SANE_FALSE;
    }
  return SANE_TRUE;
}

SANE_Byte
usb_low_io_ReadReg (ma1017 * chip, SANE_Byte RegNo)
{
  TWOPLY TP_ReadReg;
  SANE_Byte ReadByte;
  ssize_t n;

  TP_ReadReg.bByte = 0x00;
  TP_ReadReg.bRegNo = RegNo | 0x20;

  if (!chip->isOpened)
    {
      DBG (7, "usb_low_io_ReadReg Open First.\n");
      return 0xff;
    }
  if (chip->isRowing)
    {
      DBG (7, "usb_low_io_ReadReg Rowing, Stop first.\n");
      return 0xff;
    }
  if (RegNo > 0x20)
    {
      DBG (7, "usb_low_io_ReadReg RegNo out of range.\n");
      return 0xff;
    }

  n = write (chip->nUSBPipe, (void *) &TP_ReadReg, sizeof (TWOPLY));
  if (n != sizeof (TWOPLY))
    {
      DBG (7, "usb_low_io_ReadReg WriteCommand Error, fwrite return ==%d\n",
	      n);
      return 0xff;
    }
  n = read (chip->nUSBPipe, (void *) &ReadByte, 1);
  if (n != 1)
    {
      DBG (7, "usb_low_io_ReadReg Read Error!\n");
      return 0xff;
    }
  return ReadByte;
}

SANE_Bool
usb_low_io_ReadRows (ma1017 * chip, SANE_Byte * pData, SANE_Word dwByteCount)
{
  ssize_t n;
  /*
  static struct timeval last_get = { 0, 0 };
  struct timeval current;
  SANE_Word interval;
  SANE_Bool bFlag;
  */


  if (!chip->isOpened)
    {
      DBG (7, "usb_low_io_ReadRows: isOpened==SANE_FALSE\n");
      return SANE_FALSE;
    }
  if (!chip->isRowing)
    {
      DBG (7, "usb_low_io_ReadRows: isRowing==SANE_FALSE\n");
      return SANE_FALSE;
    }
  /*      usleep(1);

	bFlag = SANE_FALSE;	
	gettimeofday(&current, 0);
	if (last_get.tv_sec == 0) interval = 1000000;
	else interval = (current.tv_sec-last_get.tv_sec)*1000000+(current.tv_usec-last_get.tv_usec);
	
	if (interval<800) {
		Sleep(1);
		bFlag = SANE_TRUE;
		DBG (7, "Sleep now...,because delay == 0.\n");
	} else {
		DBG (7, "no need sleep, delayed %d us.\n",GetTickCount()-last_get);
	}
*/
  n = read (chip->nUSBPipe, (void *) pData, dwByteCount);
  /*      last_get = GetTickCount();

	        gettimeofday(&last_get,0);*/
  if (n != dwByteCount)
    {
      DBG (7, "usb_low_io_ReadRows: fread ERROR, should be %d, Get %d\n",
	      dwByteCount, n);
      if (n > 0)
	return SANE_FALSE;
      else
	return SANE_TRUE;
    }
  return SANE_TRUE;
}
